# -*- coding: utf-8 -*-
"""Pythonic wrappers for AACGM-V2 C functions that were depricated in the
change from version 2.0.0 to version 2.0.2

Functions
-------------------------------------------------------------------------------
convert : Converts array location
subsol : finds subsolar geocentric longitude and latitude
gc2gd_lat : Convert between geocentric and geodetic coordinates
igrf_dipole_axis : Get Cartesian unit vector pointing at the IGRF north dipole
------------------------------------------------------------------------------

References
-------------------------------------------------------------------------------
Laundal, K. M. and A. D. Richmond (2016), Magnetic Coordinate Systems, Space
 Sci. Rev., doi:10.1007/s11214-016-0275-y.
-------------------------------------------------------------------------------
"""

from __future__ import division, absolute_import, unicode_literals
import numpy as np
import logbook as logging

def convert(lat, lon, alt, date=None, a2g=False, trace=False, allowtrace=False,
            badidea=False, geocentric=False):
    """Converts between geomagnetic coordinates and AACGM coordinates

    Parameters
    ------------
    lat : (float)
        Input latitude in degrees N (code specifies type of latitude)
    lon : (float)
        Input longitude in degrees E (code specifies type of longitude)
    alt : (float)
        Altitude above the surface of the earth in km
    date : (datetime)
        Datetime for magnetic field
    a2g : (bool)
        True for AACGM-v2 to geographic (geodetic), False otherwise
        (default=False)
    trace : (bool)
        If True, use field-line tracing, not coefficients (default=False)
    allowtrace : (bool)
        If True, use trace only above 2000 km (default=False)
    badidea : (bool)
        If True, use coefficients above 2000 km (default=False)
    geocentric : (bool)
        True for geodetic, False for geocentric w/RE=6371.2 (default=False)

    Returns
    -------
    lat_out : (float)
        Output latitude in degrees N
    lon_out : (float)
        Output longitude in degrees E
    """
    import aacgmv2

    if(np.array(alt).max() > 2000 and not trace and not allowtrace and
       not badidea):
        estr = 'coefficients are not valid for altitudes above 2000 km. You'
        estr += ' must either use field-line tracing (trace=True '
        estr += 'or allowtrace=True) or indicate you know this is a bad idea'
        estr += ' (badidea=True)'
        logging.error(estr)
        raise ValueError(estr)

    # construct a code from the boolian flags
    bit_code = aacgmv2.convert_bool_to_bit(a2g=a2g, trace=trace,
                                           allowtrace=allowtrace,
                                           badidea=badidea,
                                           geocentric=geocentric)

    # convert location
    lat_out, lon_out, _ = aacgmv2.convert_latlon_arr(lat, lon, alt, date,
                                                     code=bit_code)

    return lat_out, lon_out

def subsol(year, doy, utime):
    """Finds subsolar geocentric longitude and latitude.

    Parameters
    ------------
    year : (int)
        Calendar year between 1601 and 2100
    doy : (int)
        Day of year between 1-365/366
    utime : (float)
        Seconds since midnight on the specified day

    Returns
    ---------
    sbsllon : (float)
        Subsolar longitude in degrees E for the given date/time
    sbsllat : (float)
        Subsolar latitude in degrees N for the given date/time

    Notes
    --------
    Based on formulas in Astronomical Almanac for the year 1996, p. C24.
    (U.S. Government Printing Office, 1994). Usable for years 1601-2100,
    inclusive. According to the Almanac, results are good to at least 0.01
    degree latitude and 0.025 degrees longitude between years 1950 and 2050.
    Accuracy for other years has not been tested. Every day is assumed to have
    exactly 86400 seconds; thus leap seconds that sometimes occur on December
    31 are ignored (their effect is below the accuracy threshold of the
    algorithm).
    After Fortran code by A. D. Richmond, NCAR. Translated from IDL
    by K. Laundal.
    """
    yr2 = year - 2000

    if year >= 2101:
        logging.error('subsol invalid after 2100. Input year is:', year)

    nleap = np.floor((year - 1601) / 4)
    nleap = nleap - 99
    if year <= 1900:
        if year <= 1600:
            print('subsol.py: subsol invalid before 1601. Input year is:', year)
        ncent = np.floor((year - 1601) / 100)
        ncent = 3 - ncent
        nleap = nleap + ncent

    l_0 = -79.549 + (-0.238699 * (yr2 - 4 * nleap) + 3.08514e-2 * nleap)
    g_0 = -2.472 + (-0.2558905 * (yr2 - 4 * nleap) - 3.79617e-2 * nleap)

    # Days (including fraction) since 12 UT on January 1 of IYR2:
    dfrac = (utime / 86400 - 1.5) + doy

    # Mean longitude of Sun:
    l_sun = l_0 + 0.9856474 * dfrac

    # Mean anomaly:
    grad = np.radians(g_0 + 0.9856003 * dfrac)

    # Ecliptic longitude:
    lmrad = np.radians(l_sun + 1.915 * np.sin(grad) + 0.020 * np.sin(2 * grad))
    sinlm = np.sin(lmrad)

    # Days (including fraction) since 12 UT on January 1 of 2000:
    epoch_day = dfrac + 365.0 * yr2 + nleap

    # Obliquity of ecliptic:
    epsrad = np.radians(23.439 - 4.0e-7 * epoch_day)

    # Right ascension:
    alpha = np.degrees(np.arctan2(np.cos(epsrad) * sinlm, np.cos(lmrad)))

    # Declination, which is the subsolar latitude:
    sbsllat = np.degrees(np.arcsin(np.sin(epsrad) * sinlm))

    # Equation of time (degrees):
    etdeg = l_sun - alpha
    etdeg = etdeg - 360.0 * np.round(etdeg / 360.0)

    # Apparent time (degrees):
    aptime = utime / 240.0 + etdeg    # Earth rotates one degree every 240 s.

    # Subsolar longitude:
    sbsllon = 180.0 - aptime
    sbsllon = sbsllon - 360.0 * np.round(sbsllon / 360.0)

    return sbsllon, sbsllat

def gc2gd_lat(gc_lat):
    """Convert geocentric latitude to geodetic latitude using WGS84.

    Parameters
    -----------
    gc_lat : (array_like or float)
        Geocentric latitude in degrees N

    Returns
    ---------
    gd_lat : (same as input)
        Geodetic latitude in degrees N
    """
    wgs84_e2 = 0.006694379990141317 - 1.0
    return np.rad2deg(-np.arctan(np.tan(np.deg2rad(gc_lat)) / wgs84_e2))

def igrf_dipole_axis(date):
    """Get Cartesian unit vector pointing at dipole pole in the north,
    according to IGRF

    Parameters
    -------------
    date : (dt.datetime)
        Date and time

    Returns
    ----------
    m_0: (np.ndarray)
        Cartesian 3 element unit vector pointing at dipole pole in the north
        (geocentric coords)

    Notes
    ----------
    IGRF coefficients are read from the igrf12coeffs.txt file. It should also
    work after IGRF updates.  The dipole coefficients are interpolated to the
    date, or extrapolated if date > latest IGRF model
    """
    import datetime as dt
    import aacgmv2

    # get time in years, as float:
    year = date.year
    doy = date.timetuple().tm_yday
    year_days = int(dt.date(date.year, 12, 31).strftime("%j"))
    year = year + doy / year_days

    # read the IGRF coefficients
    with open(aacgmv2.IGRF_COEFFS, 'r') as f_igrf:
        lines = f_igrf.readlines()

    years = lines[3].split()[3:][:-1]
    years = np.array(years, dtype=float)  # time array

    g10 = lines[4].split()[3:]
    g11 = lines[5].split()[3:]
    h11 = lines[6].split()[3:]

    # secular variation coefficients (for extrapolation)
    g10sv = np.float32(g10[-1])
    g11sv = np.float32(g11[-1])
    h11sv = np.float32(h11[-1])

    # model coefficients:
    g10 = np.array(g10[:-1], dtype=float)
    g11 = np.array(g11[:-1], dtype=float)
    h11 = np.array(h11[:-1], dtype=float)

    # get the gauss coefficient at given time:
    if year <= years[-1]:
        # regular interpolation
        g10 = np.interp(year, years, g10)
        g11 = np.interp(year, years, g11)
        h11 = np.interp(year, years, h11)
    else:
        # extrapolation
        dyear = year - years[-1]
        g10 = g10[-1] + g10sv * dyear
        g11 = g11[-1] + g11sv * dyear
        h11 = h11[-1] + h11sv * dyear

    # calculate pole position
    B_0 = np.sqrt(g10**2 + g11**2 + h11**2)

    # Calculate output
    m_0 = -np.array([g11, h11, g10]) / B_0

    return m_0



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import, unicode_literals

import numpy as np
import pytest
import aacgmv2

class TestCAACGMV2:
    def setup(self):
        """Runs before every method to create a clean testing setup"""
        self.date_args = [(2014, 3, 22, 3, 11, 0), (2018, 1, 1, 0, 0, 0)]
        self.long_date = [2014, 3, 22, 3, 11, 0]
        self.mlat = None
        self.mlon = None
        self.rshell = None
        self.mlt = None
        self.lat_in = [45.5, 60]
        self.lon_in = [-23.5, 0]
        self.alt_in = [1135, 300]

    def teardown(self):
        """Runs after every method to clean up previous testing"""
        del self.date_args, self.long_date, self.mlat, self.mlon, self.mlt
        del self.lat_in, self.lon_in, self.alt_in

    @classmethod
    def test_module_structure(self):
        """Test module structure"""
        assert aacgmv2
        assert aacgmv2._aacgmv2
        assert aacgmv2._aacgmv2.set_datetime
        assert aacgmv2._aacgmv2.convert
        assert aacgmv2._aacgmv2.inv_mlt_convert
        assert aacgmv2._aacgmv2.inv_mlt_convert_yrsec
        assert aacgmv2._aacgmv2.mlt_convert
        assert aacgmv2._aacgmv2.mlt_convert_yrsec

    def test_constants(self):
        """Test module constants"""
        ans1 = aacgmv2._aacgmv2.G2A == 0
        ans2 = aacgmv2._aacgmv2.A2G == 1
        ans3 = aacgmv2._aacgmv2.TRACE == 2
        ans4 = aacgmv2._aacgmv2.ALLOWTRACE == 4
        ans5 = aacgmv2._aacgmv2.BADIDEA == 8
        ans6 = aacgmv2._aacgmv2.GEOCENTRIC == 16

        assert ans1 & ans2 & ans3 & ans4 & ans5 & ans6
        del ans1, ans2, ans3, ans4, ans5, ans6

    def test_set_datetime(self):
        """Test set_datetime"""
        for darg in self.date_args:
            arg1 = aacgmv2._aacgmv2.set_datetime(*darg) is None
            assert arg1

    @classmethod
    def test_fail_set_datetime(self):
        """Test unsuccessful set_datetime"""
        with pytest.raises(RuntimeError):
            aacgmv2._aacgmv2.set_datetime(1013, 1, 1, 0, 0, 0)

    def test_convert_G2A_coeff(self):
        """Test convert from geographic to magnetic coordinates"""
        lat_comp = [48.1896, 58.1633]
        lon_comp = [57.7635, 81.0719]
        r_comp = [1.1775, 1.0457]

        for i,darg in enumerate(self.date_args):
            aacgmv2._aacgmv2.set_datetime(*darg)
            (self.mlat, self.mlon,
             self.rshell) = aacgmv2._aacgmv2.convert(self.lat_in[i],
                                                     self.lon_in[i],
                                                     self.alt_in[i],
                                                     aacgmv2._aacgmv2.G2A)
            np.testing.assert_almost_equal(self.mlat, lat_comp[i], decimal=4)
            np.testing.assert_almost_equal(self.mlon, lon_comp[i], decimal=4)
            np.testing.assert_almost_equal(self.rshell, r_comp[i], decimal=4)

        del lat_comp, lon_comp, r_comp

    def test_convert_A2G_coeff(self):
        """Test convert from magnetic to geodetic coordinates"""
        lat_comp = [30.7534, 50.3910]
        lon_comp = [-94.1806, -77.7919]
        r_comp = [1133.6241, 305.7138]

        for i,darg in enumerate(self.date_args):
            aacgmv2._aacgmv2.set_datetime(*darg)
            (self.mlat, self.mlon,
             self.rshell) = aacgmv2._aacgmv2.convert(self.lat_in[i],
                                                     self.lon_in[i],
                                                     self.alt_in[i],
                                                     aacgmv2._aacgmv2.A2G)
            np.testing.assert_almost_equal(self.mlat, lat_comp[i], decimal=4)
            np.testing.assert_almost_equal(self.mlon, lon_comp[i], decimal=4)
            np.testing.assert_almost_equal(self.rshell, r_comp[i], decimal=4)

        del lat_comp, lon_comp, r_comp

    def test_convert_G2A_TRACE(self):
        """Test convert from geodetic to magnetic coordinates using trace"""
        code = aacgmv2._aacgmv2.G2A + aacgmv2._aacgmv2.TRACE
        trace_lat = [48.1948, 58.1633]
        trace_lon = [57.7588, 81.0756]
        trace_r = [1.1775,  1.0457]

        for i,dargs in enumerate(self.date_args):
            aacgmv2._aacgmv2.set_datetime(*dargs)
            (self.mlat, self.mlon,
             self.rshell) = aacgmv2._aacgmv2.convert(self.lat_in[i],
                                                     self.lon_in[i],
                                                     self.alt_in[i], code)
            np.testing.assert_almost_equal(self.mlat, trace_lat[i], decimal=4)
            np.testing.assert_almost_equal(self.mlon, trace_lon[i], decimal=4)
            np.testing.assert_almost_equal(self.rshell, trace_r[i], decimal=4)

        del code, trace_lat, trace_lon, trace_r

    def test_convert_A2G_TRACE(self):
        """Test convert from magnetic to geodetic coordinates using trace"""
        code = aacgmv2._aacgmv2.A2G + aacgmv2._aacgmv2.TRACE
        trace_lat = [30.7644, 50.3958]
        trace_lon = [-94.1809, -77.8019]
        trace_r = [1133.6277, 305.7156]
        
        for i,dargs in enumerate(self.date_args):
            aacgmv2._aacgmv2.set_datetime(*dargs)
            (self.mlat, self.mlon,
             self.rshell) = aacgmv2._aacgmv2.convert(self.lat_in[i],
                                                     self.lon_in[i],
                                                     self.alt_in[i], code)
            np.testing.assert_almost_equal(self.mlat, trace_lat[i], decimal=4)
            np.testing.assert_almost_equal(self.mlon, trace_lon[i], decimal=4)
            np.testing.assert_almost_equal(self.rshell, trace_r[i], decimal=4)

        del code, trace_lat, trace_lon, trace_r

    def test_convert_high_denied(self):
        """Test for failure when converting to high altitude geodetic to
        magnetic coordinates"""
        aacgmv2._aacgmv2.set_datetime(*self.date_args[0])
        with pytest.raises(RuntimeError):
            aacgmv2._aacgmv2.convert(self.lat_in[0], self.lon_in[0], 5500,
                                     aacgmv2._aacgmv2.G2A)

    def test_convert_high_TRACE(self):
        """Test convert from high altitude geodetic to magnetic coordinates
        using trace"""
        code = aacgmv2._aacgmv2.G2A + aacgmv2._aacgmv2.TRACE
        aacgmv2._aacgmv2.set_datetime(*self.date_args[0])
        (self.mlat, self.mlon,
         self.rshell) = aacgmv2._aacgmv2.convert(self.lat_in[0], self.lon_in[0],
                                                 5500, code)
        np.testing.assert_almost_equal(self.mlat, 59.9748, decimal=4)
        np.testing.assert_almost_equal(self.mlon, 57.7425, decimal=4)
        np.testing.assert_almost_equal(self.rshell, 1.8626, decimal=4)

        del code

    def test_convert_high_ALLOWTRACE(self):
        """Test convert from high altitude geodetic to magnetic coordinates
        by allowing IGRF tracing"""
        code = aacgmv2._aacgmv2.G2A + aacgmv2._aacgmv2.ALLOWTRACE
        aacgmv2._aacgmv2.set_datetime(*self.date_args[0])
        (self.mlat, self.mlon,
         self.rshell) = aacgmv2._aacgmv2.convert(self.lat_in[0], self.lon_in[0],
                                                 5500, code)
        np.testing.assert_almost_equal(self.mlat, 59.9748, decimal=4)
        np.testing.assert_almost_equal(self.mlon, 57.7425, decimal=4)
        np.testing.assert_almost_equal(self.rshell, 1.8626, decimal=4)

        del code

    def test_convert_high_BADIDEA(self):
        """Test convert from high altitude geodetic to magnetic coordinates
        using coefficients"""
        code = aacgmv2._aacgmv2.G2A + aacgmv2._aacgmv2.BADIDEA
        aacgmv2._aacgmv2.set_datetime(*self.date_args[0])
        (self.mlat, self.mlon,
         self.rshell) = aacgmv2._aacgmv2.convert(self.lat_in[0], self.lon_in[0],
                                                 5500, code)
        np.testing.assert_almost_equal(self.mlat, 58.7154, decimal=4)
        np.testing.assert_almost_equal(self.mlon, 56.5830, decimal=4)
        np.testing.assert_almost_equal(self.rshell, 1.8626, decimal=4)

        del code

    def test_convert_GEOCENTRIC_G2A_coeff(self):
        """Test convert from geographic to magnetic coordinates"""
        code = aacgmv2._aacgmv2.G2A + aacgmv2._aacgmv2.GEOCENTRIC
        aacgmv2._aacgmv2.set_datetime(*self.date_args[0])
        (self.mlat, self.mlon,
         self.rshell) = aacgmv2._aacgmv2.convert(self.lat_in[0], self.lon_in[0],
                                                 self.alt_in[0], code)
        np.testing.assert_almost_equal(self.mlat, 48.3779, decimal=4)
        np.testing.assert_almost_equal(self.mlon, 57.7974, decimal=4)
        np.testing.assert_almost_equal(self.rshell, 1.1781, decimal=4)

        del code
        
    def test_convert_GEOCENTRIC_A2G_coeff(self):
        """Test convert from magnetic to geocentric coordinates"""
        code = aacgmv2._aacgmv2.A2G + aacgmv2._aacgmv2.GEOCENTRIC
        aacgmv2._aacgmv2.set_datetime(*self.date_args[0])
        (self.mlat, self.mlon,
         self.rshell) = aacgmv2._aacgmv2.convert(self.lat_in[0], self.lon_in[0],
                                                 self.alt_in[0], code)
        np.testing.assert_almost_equal(self.mlat, 30.6101, decimal=4)
        np.testing.assert_almost_equal(self.mlon, -94.1806, decimal=4)
        np.testing.assert_almost_equal(self.rshell, 1135.0000, decimal=4)

        del code

    def test_convert_GEOCENTRIC_G2A_TRACE(self):
        """Test convert from geographic to magnetic coordinates using trace"""
        code = aacgmv2._aacgmv2.G2A + aacgmv2._aacgmv2.TRACE + \
               aacgmv2._aacgmv2.GEOCENTRIC
        aacgmv2._aacgmv2.set_datetime(*self.date_args[0])
        (self.mlat, self.mlon,
         self.rshell) = aacgmv2._aacgmv2.convert(self.lat_in[0], self.lon_in[0],
                                                 self.alt_in[0], code)
        np.testing.assert_almost_equal(self.mlat, 48.3830, decimal=4)
        np.testing.assert_almost_equal(self.mlon, 57.7926, decimal=4)
        np.testing.assert_almost_equal(self.rshell, 1.1781, decimal=4)

        del code

    def test_convert_GEOCENTRIC_A2G_TRACE(self):
        """Test convert from magnetic to geographic coordinates using trace"""
        code = aacgmv2._aacgmv2.A2G + aacgmv2._aacgmv2.TRACE + \
               aacgmv2._aacgmv2.GEOCENTRIC
        aacgmv2._aacgmv2.set_datetime(*self.date_args[0])
        (self.mlat, self.mlon,
         self.rshell) = aacgmv2._aacgmv2.convert(self.lat_in[0], self.lon_in[0],
                                                 self.alt_in[0], code)
        np.testing.assert_almost_equal(self.mlat, 30.6211, decimal=4)
        np.testing.assert_almost_equal(self.mlon, -94.1809, decimal=4)
        np.testing.assert_almost_equal(self.rshell, 1135.0000, decimal=4)

        del code

    @classmethod
    def test_forbidden(self):
        """Test convert failure"""
        with pytest.raises(RuntimeError):
            aacgmv2._aacgmv2.convert(7, 0, 0, aacgmv2._aacgmv2.G2A)

    def test_inv_mlt_convert(self):
        """Test MLT inversion"""
        mlt_args = list(self.long_date)
        mlt_args.extend([12.0])
        self.mlon = aacgmv2._aacgmv2.inv_mlt_convert(*mlt_args)
        np.testing.assert_almost_equal(self.mlon, -153.5931, decimal=4)

        mlt_args[-1] = 25.0
        self.mlon = aacgmv2._aacgmv2.inv_mlt_convert(*mlt_args)
        np.testing.assert_almost_equal(self.mlon, 41.4069, decimal=4)

        mlt_args[-1] = -1.0
        self.mlon = aacgmv2._aacgmv2.inv_mlt_convert(*mlt_args)
        np.testing.assert_almost_equal(self.mlon, 11.4069, decimal=4)

        del mlt_args

    def test_inv_mlt_convert_yrsec(self):
        """Test MLT inversion with year and seconds of year"""
        import datetime as dt
        dtime = dt.datetime(*self.long_date)
        soy = (int(dtime.strftime("%j"))-1) * 86400 + dtime.hour * 3600 + \
              dtime.minute * 60 + dtime.second
        
        mlt_args_1 = [dtime.year, soy, 12.0]
        mlt_args_2 = [dtime.year, soy, 25.0]
        mlt_args_3 = [dtime.year, soy, -1.0]

        mlon_1 = aacgmv2._aacgmv2.inv_mlt_convert_yrsec(*mlt_args_1)
        mlon_2 = aacgmv2._aacgmv2.inv_mlt_convert_yrsec(*mlt_args_2)
        mlon_3 = aacgmv2._aacgmv2.inv_mlt_convert_yrsec(*mlt_args_3)

        np.testing.assert_almost_equal(mlon_1, -153.5931, decimal=4)
        np.testing.assert_almost_equal(mlon_2, 41.4069, decimal=4)
        np.testing.assert_almost_equal(mlon_3, 11.4069, decimal=4)

        del dtime, soy, mlt_args_1, mlt_args_2, mlt_args_3, mlon_1, mlon_2
        del mlon_3

    def test_mlt_convert(self):
        """Test MLT calculation with different longitudes"""
        mlt_args = list(self.long_date)
        mlt_args.extend([270.0])
        self.mlt = aacgmv2._aacgmv2.mlt_convert(*mlt_args)
        np.testing.assert_almost_equal(self.mlt, 16.2395, decimal=4)

        mlt_args[-1] = 80.0
        self.mlt = aacgmv2._aacgmv2.mlt_convert(*mlt_args)
        np.testing.assert_almost_equal(self.mlt, 3.5729, decimal=4)

        mlt_args[-1] = -90.0
        self.mlt = aacgmv2._aacgmv2.mlt_convert(*mlt_args)
        np.testing.assert_almost_equal(self.mlt, 16.2395, decimal=4)

        del mlt_args

    def test_mlt_convert_yrsec(self):
        """Test MLT calculation using year and seconds of year"""
        import datetime as dt
        dtime = dt.datetime(*self.long_date)
        soy = (int(dtime.strftime("%j"))-1) * 86400 + dtime.hour * 3600 + \
            dtime.minute * 60 + dtime.second
        mlt_args_1 = [dtime.year, soy, 270.0]
        mlt_args_2 = [dtime.year, soy, 80.0]
        mlt_args_3 = [dtime.year, soy, -90.0]
        
        mlt_1 = aacgmv2._aacgmv2.mlt_convert_yrsec(*mlt_args_1)
        mlt_2 = aacgmv2._aacgmv2.mlt_convert_yrsec(*mlt_args_2)
        mlt_3 = aacgmv2._aacgmv2.mlt_convert_yrsec(*mlt_args_3)

        np.testing.assert_almost_equal(mlt_1, 16.2395, decimal=4)
        np.testing.assert_almost_equal(mlt_2, 3.5729, decimal=4)
        np.testing.assert_equal(mlt_1, mlt_3)

        del dtime, soy, mlt_args_1, mlt_args_2, mlt_args_3, mlt_1, mlt_2, mlt_3



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinxcontrib.napoleon',
    'numpydoc'
]
if os.getenv('SPELLCHECK'):
    extensions += 'sphinxcontrib.spelling',
    spelling_show_suggestions = True
    spelling_lang = 'en_US'

source_suffix = '.rst'
master_doc = 'index'
project = u'AACGM-v2 Python library'
year = u'2015'
author = u'Angeline G. Burrell, Christer van der Meeren'
copyright = '{0}, {1}'.format(year, author)
version = release = u'2.5.0'
# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

# only import and set the theme if we're building docs locally
if not on_rtd:
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

pygments_style = 'trac'
templates_path = ['.']
html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_split_index = True
html_sidebars = {
   '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
}
html_short_title = '%s-%s' % (project, version)



# -*- coding: utf-8 -*-


"""Executed when aacgmv2 is invoked with python -m aacgmv2"""

from __future__ import division, print_function, absolute_import

import sys
import argparse
import datetime as dt
import numpy as np
import aacgmv2

try:
    # Python 3
    STDIN = sys.stdin.buffer
    STDOUT = sys.stdout.buffer
except AttributeError:
    # Python 2
    STDIN = sys.stdin
    STDOUT = sys.stdout


def main():
    """Entry point for the script"""

    desc = 'Converts between geographical coordinates, AACGM-v2, and MLT'
    parser = argparse.ArgumentParser(description=desc)

    desc = 'for help, run %(prog)s SUBCOMMAND -h'
    subparsers = parser.add_subparsers(title='Subcommands', prog='aacgmv2',
                                       dest='subcommand', description=desc)
    subparsers.required = True

    desc = 'convert to/from geomagnetic coordinates. Input file must have lines'
    desc += 'of the form "LAT LON ALT".'
    parser_convert = subparsers.add_parser('convert', help=(desc))

    desc = 'convert between magnetic local time (MLT) and AACGM-v2 longitude. '
    desc += 'Input file must have a single number on each line.'
    parser_convert_mlt = subparsers.add_parser('convert_mlt', help=(desc))

    desc = 'input file (stdin if none specified)'
    for pp in [parser_convert, parser_convert_mlt]:
        pp.add_argument('-i', '--input', dest='file_in', metavar='FILE_IN',
                        type=argparse.FileType('r'), default=STDIN, help=desc)
        pp.add_argument('-o', '--output', dest='file_out', metavar='FILE_OUT',
                        type=argparse.FileType('wb'), default=STDOUT,
                        help='output file (stdout if none specified)')

    desc = 'date for magnetic field model (1900-2020, default: today)'
    parser_convert.add_argument('-d', '--date', dest='date', metavar='YYYYMMDD',
                                help=desc)

    desc = 'invert - convert AACGM to geographic instead of geographic to AACGM'
    parser_convert.add_argument('-v', '--a2g', dest='a2g', action='store_true',
                                default=False, help=desc)

    desc = 'use field-line tracing instead of coefficients'
    parser_convert.add_argument('-t', '--trace', dest='trace',
                                action='store_true', default=False, help=desc)

    desc = 'automatically use field-line tracing above 2000 km'
    parser_convert.add_argument('-a', '--allowtrace', dest='allowtrace',
                                action='store_true', default=False, help=desc)

    desc = 'allow use of coefficients above 2000 km (bad idea!)'
    parser_convert.add_argument('-b', '--badidea', dest='badidea',
                                action='store_true', default=False, help=desc)

    desc = 'assume inputs are geocentric with Earth radius 6371.2 km'
    parser_convert.add_argument('-g', '--geocentric', dest='geocentric',
                                action='store_true', default=False, help=desc)

    parser_convert_mlt.add_argument('datetime', metavar='YYYYMMDDHHMMSS',
                                    help='date and time for conversion')

    desc = 'invert - convert MLT to AACGM longitude instead of AACGM longitude'
    desc += ' to MLT'
    parser_convert_mlt.add_argument('-v', '--m2a', dest='m2a',
                                    action='store_true', default=False,
                                    help=desc)

    args = parser.parse_args()
    array = np.loadtxt(args.file_in, ndmin=2)

    if args.subcommand == 'convert':
        date = dt.date.today() if args.date is None else \
               dt.datetime.strptime(args.date, '%Y%m%d')
        code = aacgmv2.convert_bool_to_bit(a2g=args.a2g, trace=args.trace,
                                           allowtrace=args.allowtrace,
                                           badidea=args.badidea,
                                           geocentric=args.geocentric)
        lats, lons, alts = aacgmv2.convert_latlon_arr(array[:, 0], array[:, 1],
                                                      array[:, 2], dtime=date,
                                                      code=code)
        np.savetxt(args.file_out, np.column_stack((lats, lons, alts)),
                   fmt='%.8f')
    elif args.subcommand == 'convert_mlt':
        dtime = dt.datetime.strptime(args.datetime, '%Y%m%d%H%M%S')
        out = aacgmv2.convert_mlt(array[:, 0], dtime, m2a=args.m2a)
        np.savetxt(args.file_out, out, fmt='%.8f')


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import, unicode_literals

import subprocess
import numpy as np

class testCmdAACGMV2:
    def setup(self):
        """Runs before every method to create a clean testing setup"""
        self.output = "tests/test_data/output.txt"
        self.convert = "tests/test_data/test_convert.txt"
        self.single = 'tests/test_data/test_convert_single_line.txt'
        self.mlt = 'tests/test_data/test_convert_mlt.txt'
        self.mlt_single = 'tests/test_data/test_convert_mlt_single_line.txt'

    def teardown(self):
        """Runs after every method to clean up previous testing"""
        import os

        if os.path.isfile(self.output):
            os.remove(self.output)

        del self.output, self.convert, self.single, self.mlt, self.mlt_single

    def test_module_invocation(self):
        p = subprocess.Popen(['python', '-m', 'aacgmv2', 'convert', '-i',
                              'tests/test_convert.txt', '-d', '20150224',
                              '-o', self.output])
        p.communicate()
        p.wait()
        data = np.loadtxt(self.output)
        np.testing.assert_allclose(data, [[57.4761, 93.5572],
                                          [58.5332, 93.9607],
                                          [59.5852, 94.3897]], rtol=1e-4)

    def test_convert_g2a(self):
        p = subprocess.Popen(['aacgmv2', 'convert', '-i', self.convert, '-d',
                              '20150224', '-o', self.output])
        p.communicate()
        p.wait()
        data = np.loadtxt(self.output)
        np.testing.assert_allclose(data, [[57.4761, 93.5572],
                                          [58.5332, 93.9607],
                                          [59.5852, 94.3897]], rtol=1e-4)

    def test_convert_a2g(self):
        p = subprocess.Popen(['aacgmv2', 'convert', '-i', self.convert,
                              '-d', '20150224', '-o', self.output, '-v'])
        p.communicate()
        p.wait()
        data = np.loadtxt(self.output)
        np.testing.assert_allclose(data, [[51.6547, -66.6601],
                                          [52.6725, -66.7555],
                                          [53.6914, -66.8552]], rtol=1e-4)

    def test_convert_trace_g2a(self):
        p = subprocess.Popen(['aacgmv2', 'convert', '-i', self.convert, '-d',
                              '20150224', '-o', self.output, '-t'])
        p.communicate()
        p.wait()
        data = np.loadtxt(self.output)
        np.testing.assert_allclose(data, [[57.4736, 93.5676],
                                          [58.5305, 93.9716],
                                          [59.5825, 94.4009]], rtol=1e-4)

    def test_convert_trace_a2g(self):
        p = subprocess.Popen(['aacgmv2', 'convert', '-i', self.convert, '-d',
                              '20150224', '-o', self.output, '-t', '-v'])
        p.communicate()
        p.wait()
        data = np.loadtxt(self.output)
        np.testing.assert_allclose(data, [[51.6454, -66.6444],
                                          [52.6671, -66.7432],
                                          [53.6899, -66.8469]], rtol=1e-4)

    def test_convert_geocentric(self):
        p = subprocess.Popen(['aacgmv2', 'convert', '-i', self.convert, '-d',
                              '20150224', '-o', self.output, '-g'])
        p.communicate()
        p.wait()
        data = np.loadtxt(self.output)
        np.testing.assert_allclose(data, [[57.6697, 93.6319],
                                          [58.7223, 94.0385],
                                          [59.7695, 94.4708]], rtol=1e-4)

    def test_convert_today(self):
        p = subprocess.Popen(['aacgmv2', 'convert', '-i', self.convert])
        p.communicate()
        p.wait()

    def test_convert_single_line(self):
        p = subprocess.Popen(['aacgmv2', 'convert', '-i', self.single, '-d',
                              '20150224', '-o', self.output])
        p.communicate()
        p.wait()
        data = np.loadtxt(self.output)
        np.testing.assert_allclose(data, [57.4761, 93.5572], rtol=1e-4)

    def test_convert_stdin_stdout(self):
        p = subprocess.Popen('echo 60 15 300 | aacgmv2 convert -d 20150224',
                             shell=True, stdout=subprocess.PIPE)
        stdout, _ = p.communicate()
        p.wait()
        assert b'57.47612194 93.55719875' in stdout

    def test_convert_mlt_a2m(self):
        p = subprocess.Popen(['aacgmv2', 'convert_mlt', '-i', self.mlt,
                              '20150224140015', '-o', self.output])
        p.communicate()
        p.wait()
        data = np.loadtxt(self.output)
        np.testing.assert_allclose(data, [9.056476, 9.78981, 10.523143],
                                   rtol=1e-6)

    def test_convert_mlt_m2a(self):
        p = subprocess.Popen(['aacgmv2', 'convert_mlt', '-i', self.mlt,
                              '20150224140015', '-o', self.output, '-v'])
        p.communicate()
        p.wait()
        data = np.loadtxt(self.output)
        np.testing.assert_allclose(data, [240.152854, 45.152854, 210.152854],
                                   rtol=1e-6)

    def test_convert_mlt_single_line(self):
        p = subprocess.Popen(['aacgmv2', 'convert_mlt', '-i', self.mlt_single,
                              '20150224140015', '-o', self.output])
        p.communicate()
        p.wait()
        data = np.loadtxt(self.output)
        np.testing.assert_allclose(data, 9.0564764, rtol=1e-6)

    @classmethod
    def test_convert_mlt_stdin_stdout(self):
        p = subprocess.Popen('echo 12 | aacgmv2 convert_mlt -v 20150224140015',
                             shell=True, stdout=subprocess.PIPE)
        stdout, _ = p.communicate()
        p.wait()
        assert b'45.15285362' in stdout

    def test_convert_mlt_stdin_stdout_order(self):
        p = subprocess.Popen('echo 12 | aacgmv2 convert_mlt 20150224140015 -v',
                             shell=True, stdout=subprocess.PIPE)
        stdout, _ = p.communicate()
        p.wait()
        assert b'45.15285362' in stdout



# -*- coding: utf-8 -*-
"""Pythonic wrappers for AACGM-V2 C functions.

Functions
--------------
set_coeff_path : Set the coefficient paths using default or supplied values
convert_latlon : Converts scalar location
convert_latlon_arr : Converts array location
get_aacgm_coord : Get scalar magnetic lat, lon, mlt from geographic location
get_aacgm_coord_arr : Get array magnetic lat, lon, mlt from geographic location
convert_str_to_bit : Convert human readible AACGM flag to bits
convert_bool_to_bit : Convert boolian flags to bits
convert_mlt : Get array mlt
--------------
"""

from __future__ import division, absolute_import, unicode_literals
import datetime as dt
import numpy as np
import logbook as logging

def set_coeff_path(igrf_file=False, coeff_prefix=False):
    """Sets the IGRF_COEFF and AACGMV_V2_DAT_PREFIX environment variables.

    Parameters
    -----------
    igrf_file : (str or bool)
        Full filename of IGRF coefficient file, True to use
        aacgmv2.IGRF_COEFFS, or False to leave as is. (default=False)
    coeff_prefix : (str or bool)
        Location and file prefix for aacgm coefficient files, True to use
        aacgmv2.AACGM_V2_DAT_PREFIX, or False to leave as is. (default=False)

    Returns
    ---------
    Void
    """
    import aacgmv2
    import os

    # Define coefficient file prefix if requested
    if coeff_prefix is not False:
        # Use the default value, if one was not supplied (allow None to
        # comply with depricated behaviour)
        if coeff_prefix is True or coeff_prefix is None:
            coeff_prefix = aacgmv2.AACGM_v2_DAT_PREFIX

        if hasattr(os, "unsetenv"):
            os.unsetenv('AACGM_v2_DAT_PREFIX')
        else:
            del os.environ['AACGM_v2_DAT_PREFIX']
        os.environ['AACGM_v2_DAT_PREFIX'] = coeff_prefix

    # Define IGRF file if requested
    if igrf_file is not False:
        # Use the default value, if one was not supplied (allow None to
        # comply with depricated behaviour)
        if igrf_file is True or igrf_file is None:
            igrf_file = aacgmv2.IGRF_COEFFS

        if hasattr(os, "unsetenv"):
            os.unsetenv('IGRF_COEFFS')
        else:
            del os.environ['IGRF_COEFFS']
        os.environ['IGRF_COEFFS'] = igrf_file

    return

def convert_latlon(in_lat, in_lon, height, dtime, code="G2A"):
    """Converts between geomagnetic coordinates and AACGM coordinates

    Parameters
    ------------
    in_lat : (float)
        Input latitude in degrees N (code specifies type of latitude)
    in_lon : (float)
        Input longitude in degrees E (code specifies type of longitude)
    height : (float)
        Altitude above the surface of the earth in km
    dtime : (datetime)
        Datetime for magnetic field
    code : (str or int)
        Bit code or string denoting which type(s) of conversion to perform
        G2A        - geographic (geodetic) to AACGM-v2
        A2G        - AACGM-v2 to geographic (geodetic)
        TRACE      - use field-line tracing, not coefficients
        ALLOWTRACE - use trace only above 2000 km
        BADIDEA    - use coefficients above 2000 km
        GEOCENTRIC - assume inputs are geocentric w/ RE=6371.2
        (default is "G2A")

    Returns
    -------
    out_lat : (float)
        Output latitude in degrees N
    out_lon : (float)
        Output longitude in degrees E
    out_r : (float)
        Geocentric radial distance (R_Earth) or altitude above the surface of
        the Earth (km)
    """
    import aacgmv2._aacgmv2 as c_aacgmv2

    # Test time
    if isinstance(dtime, dt.date):
        dtime = dt.datetime.combine(dtime, dt.time(0))

    assert isinstance(dtime, dt.datetime), \
        logging.error('time must be specified as datetime object')

    # Test height
    if height < 0:
        logging.warn('conversion not intended for altitudes < 0 km')

    # Initialise output
    lat_out = np.nan
    lon_out = np.nan
    r_out = np.nan

    # Test code
    try:
        code = code.upper()

        if(height > 2000 and code.find("TRACE") < 0 and
           code.find("ALLOWTRACE") < 0 and code.find("BADIDEA") < 0):
            estr = 'coefficients are not valid for altitudes above 2000 km. You'
            estr += ' must either use field-line tracing (trace=True '
            estr += 'or allowtrace=True) or indicate you know this '
            estr += 'is a bad idea'
            logging.error(estr)
            return lat_out, lon_out, r_out

        # make flag
        bit_code = convert_str_to_bit(code)
    except AttributeError:
        bit_code = code

    assert isinstance(bit_code, int), \
        logging.error("unknown code {:}".format(bit_code))

    # Test latitude range
    if abs(in_lat) > 90.0:
        assert abs(in_lat) <= 90.1, logging.error('unrealistic latitude')
        in_lat = np.sign(in_lat) * 90.0

    # Constrain longitudes between -180 and 180
    in_lon = ((in_lon + 180.0) % 360.0) - 180.0

    # Set current date and time
    try:
        c_aacgmv2.set_datetime(dtime.year, dtime.month, dtime.day, dtime.hour,
                               dtime.minute, dtime.second)
    except:
        raise RuntimeError("unable to set time for {:}".format(dtime))

    # convert location
    try:
        lat_out, lon_out, r_out = c_aacgmv2.convert(in_lat, in_lon, height,
                                                    bit_code)
    except:
        pass

    return lat_out, lon_out, r_out

def convert_latlon_arr(in_lat, in_lon, height, dtime, code="G2A"):
    """Converts between geomagnetic coordinates and AACGM coordinates.

    Parameters
    ------------
    in_lat : (np.ndarray or list or float)
        Input latitude in degrees N (code specifies type of latitude)
    in_lon : (np.ndarray or list or float)
        Input longitude in degrees E (code specifies type of longitude)
    height : (np.ndarray or list or float)
        Altitude above the surface of the earth in km
    dtime : (datetime)
        Single datetime object for magnetic field
    code : (int or str)
        Bit code or string denoting which type(s) of conversion to perform
        G2A        - geographic (geodetic) to AACGM-v2
        A2G        - AACGM-v2 to geographic (geodetic)
        TRACE      - use field-line tracing, not coefficients
        ALLOWTRACE - use trace only above 2000 km
        BADIDEA    - use coefficients above 2000 km
        GEOCENTRIC - assume inputs are geocentric w/ RE=6371.2
        (default = "G2A")

    Returns
    -------
    out_lat : (np.ndarray)
        Output latitudes in degrees N
    out_lon : (np.ndarray)
        Output longitudes in degrees E
    out_r : (np.ndarray)
        Geocentric radial distance (R_Earth) or altitude above the surface of
        the Earth (km)

    Notes
    -------
    At least one of in_lat, in_lon, and height must be a list or array.
    """
    import aacgmv2._aacgmv2 as c_aacgmv2

    # If a list was entered instead of a numpy array, recast it here
    if isinstance(in_lat, list):
        in_lat = np.array(in_lat)

    if isinstance(in_lon, list):
        in_lon = np.array(in_lon)

    if isinstance(height, list):
        height = np.array(height)

    # If one or two of these elements is a float or int, create an array
    test_array = np.array([hasattr(in_lat, "shape"), hasattr(in_lon, "shape"),
                           hasattr(height, "shape")])
    if not test_array.all():
        if test_array.any():
            arr_shape = in_lat.shape if test_array.argmax() == 0 else \
                        (in_lon.shape if test_array.argmax() == 1 else
                         height.shape)
            if not test_array[0]:
                in_lat = np.ones(shape=arr_shape, dtype=float) * in_lat
            if not test_array[1]:
                in_lon = np.ones(shape=arr_shape, dtype=float) * in_lon
            if not test_array[2]:
                height = np.ones(shape=arr_shape, dtype=float) * height
        else:
            logging.info("for a single location, consider using convert_latlon")
            in_lat = np.array([in_lat])
            in_lon = np.array([in_lon])
            height = np.array([height])

    # Ensure that lat, lon, and height are the same length or if the lengths
    # differ that the different ones contain only a single value
    if not (in_lat.shape == in_lon.shape and in_lat.shape == height.shape):
        ulen = np.unique([in_lat.shape, in_lon.shape, height.shape])
        if ulen.min() != (1,):
            logging.error("mismatched input arrays")
            return None, None, None

    # Test time
    if isinstance(dtime, dt.date):
        dtime = dt.datetime.combine(dtime, dt.time(0))

    assert isinstance(dtime, dt.datetime), \
        logging.error('time must be specified as datetime object')

    # Test height
    if np.min(height) < 0:
        logging.warn('conversion not intended for altitudes < 0 km')

    # Initialise output
    lat_out = np.empty(shape=in_lat.shape, dtype=float) * np.nan
    lon_out = np.empty(shape=in_lon.shape, dtype=float) * np.nan
    r_out = np.empty(shape=height.shape, dtype=float) * np.nan

    # Test code
    try:
        code = code.upper()

        if(np.nanmax(height) > 2000 and code.find("TRACE") < 0 and
           code.find("ALLOWTRACE") < 0 and code.find("BADIDEA") < 0):
            estr = 'coefficients are not valid for altitudes above 2000 km. You'
            estr += ' must either use field-line tracing (trace=True '
            estr += 'or allowtrace=True) or indicate you know this '
            estr += 'is a bad idea'
            logging.error(estr)
            return lat_out, lon_out, r_out

        # make flag
        bit_code = convert_str_to_bit(code)
    except AttributeError:
        bit_code = code

    assert isinstance(bit_code, int), \
        logging.error("unknown code {:}".format(bit_code))

    # Test latitude range
    if np.abs(in_lat).max() > 90.0:
        assert np.abs(in_lat).max() <= 90.1, \
            logging.error('unrealistic latitude')
        in_lat = np.clip(in_lat, -90.0, 90.0)

    # Constrain longitudes between -180 and 180
    in_lon = ((in_lon + 180.0) % 360.0) - 180.0

    # Set current date and time
    try:
        c_aacgmv2.set_datetime(dtime.year, dtime.month, dtime.day, dtime.hour,
                               dtime.minute, dtime.second)
    except:
        raise RuntimeError("unable to set time for {:}".format(dtime))

    # Vectorise the AACGM code
    convert_vectorised = np.vectorize(c_aacgmv2.convert)

    # convert
    try:
        lat_out, lon_out, r_out = convert_vectorised(in_lat, in_lon, height,
                                                     bit_code)
    except:
        pass

    return lat_out, lon_out, r_out

def get_aacgm_coord(glat, glon, height, dtime, method="TRACE"):
    """Get AACGM latitude, longitude, and magnetic local time

    Parameters
    ------------
    glat : (float)
        Geodetic latitude in degrees N
    glon : (float)
        Geodetic longitude in degrees E
    height : (float)
        Altitude above the surface of the earth in km
    dtime : (datetime)
        Date and time to calculate magnetic location
    method : (str)
        String denoting which type(s) of conversion to perform
        TRACE      - use field-line tracing, not coefficients
        ALLOWTRACE - use trace only above 2000 km
        BADIDEA    - use coefficients above 2000 km
        GEOCENTRIC - assume inputs are geocentric w/ RE=6371.2
        (default = "TRACE")

    Returns
    -------
    mlat : (float)
        magnetic latitude in degrees N
    mlon : (float)
        magnetic longitude in degrees E
    mlt : (float)
        magnetic local time in hours
    """
    # Initialize code
    code = "G2A|{:s}".format(method)

    # Get magnetic lat and lon.
    mlat, mlon, _ = convert_latlon(glat, glon, height, dtime, code=code)

    # Get magnetic local time
    if np.isnan(mlon):
        mlt = np.nan
    else:
        mlt = convert_mlt(mlon, dtime, m2a=False)

    return mlat, mlon, mlt


def get_aacgm_coord_arr(glat, glon, height, dtime, method="TRACE"):
    """Get AACGM latitude, longitude, and magnetic local time

    Parameters
    ------------
    glat : (np.array or list)
        Geodetic latitude in degrees N
    glon : (np.array or list)
        Geodetic longitude in degrees E
    height : (np.array or list)
        Altitude above the surface of the earth in km
    dtime : (datetime)
        Date and time to calculate magnetic location
    method : (str)
        String denoting which type(s) of conversion to perform
        TRACE      - use field-line tracing, not coefficients
        ALLOWTRACE - use trace only above 2000 km
        BADIDEA    - use coefficients above 2000 km
        GEOCENTRIC - assume inputs are geocentric w/ RE=6371.2
        (default = "TRACE")
        (default = "TRACE")

    Returns
    -------
    mlat : (float)
        magnetic latitude in degrees N
    mlon : (float)
        magnetic longitude in degrees E
    mlt : (float)
        magnetic local time in hours
    """
    # Initialize code
    code = "G2A|{:s}".format(method)

    # Get magnetic lat and lon.
    mlat, mlon, _ = convert_latlon_arr(glat, glon, height, dtime, code=code)

    if np.all(np.isnan(mlon)):
        mlt = np.empty(shape=mlat.shape, dtype=float) * np.nan
    else:
        # Get magnetic local time
        mlt = convert_mlt(mlon, dtime, m2a=False)

    return mlat, mlon, mlt

def convert_str_to_bit(code):
    """convert string code specification to bit code specification

    Parameters
    ------------
    code : (str)
        Bitwise code for passing options into converter (default=0)
        G2A        - geographic (geodetic) to AACGM-v2
        A2G        - AACGM-v2 to geographic (geodetic)
        TRACE      - use field-line tracing, not coefficients
        ALLOWTRACE - use trace only above 2000 km
        BADIDEA    - use coefficients above 2000 km
        GEOCENTRIC - assume inputs are geocentric w/ RE=6371.2

    Returns
    --------
    bit_code : (int)
        code specification in bits

    Notes
    --------
    Multiple codes should be seperated by pipes '|'.  Invalid parts of the code
    are ignored and no code defaults to 'G2A'.
    """
    import aacgmv2._aacgmv2 as c_aacgmv2

    convert_code = {"G2A": c_aacgmv2.G2A, "A2G": c_aacgmv2.A2G,
                    "TRACE": c_aacgmv2.TRACE, "BADIDEA": c_aacgmv2.BADIDEA,
                    "GEOCENTRIC": c_aacgmv2.GEOCENTRIC,
                    "ALLOWTRACE": c_aacgmv2.ALLOWTRACE}

    # Force upper case, remove any spaces, and split along pipes
    codes = code.upper().replace(" ", "").split("|")

    # Add the valid parts of the code, invalid elements are ignored
    bit_code = sum([convert_code[k] for k in codes if k in convert_code.keys()])

    return bit_code

def convert_bool_to_bit(a2g=False, trace=False, allowtrace=False,
                        badidea=False, geocentric=False):
    """convert boolian flags to bit code specification

    Parameters
    ----------
    a2g : (bool)
        True for AACGM-v2 to geographic (geodetic), False otherwise
        (default=False)
    trace : (bool)
        If True, use field-line tracing, not coefficients (default=False)
    allowtrace : (bool)
        If True, use trace only above 2000 km (default=False)
    badidea : (bool)
        If True, use coefficients above 2000 km (default=False)
    geocentric : (bool)
        True for geodetic, False for geocentric w/RE=6371.2 (default=False)

    Returns
    --------
    bit_code : (int)
        code specification in bits
    """
    import aacgmv2._aacgmv2 as c_aacgmv2

    bit_code = c_aacgmv2.A2G if a2g else c_aacgmv2.G2A

    if trace:
        bit_code += c_aacgmv2.TRACE
    if allowtrace:
        bit_code += c_aacgmv2.ALLOWTRACE
    if badidea:
        bit_code += c_aacgmv2.BADIDEA
    if geocentric:
        bit_code += c_aacgmv2.GEOCENTRIC

    return bit_code

def convert_mlt(arr, dtime, m2a=False):
    """Converts between magnetic local time (MLT) and AACGM-v2 longitude

    Parameters
    ------------
    arr : (array_line or float)
        Magnetic longitudes (degrees E) or MLTs (hours) to convert
    dtime : (datetime.datetime)
        Date and time for MLT conversion in Universal Time (UT).
    m2a : (bool)
        Convert MLT to AACGM-v2 longitude (True) or magnetic longitude to MLT
        (False).  (default=False)

    Returns
    --------
    out : (np.ndarray)
        Converted coordinates/MLT in degrees E or hours (as appropriate)

    Notes
    -------
    This routine previously based on Laundal et al. 2016, but now uses the
    improved calculation available in AACGM-V2.4.
    """
    import aacgmv2._aacgmv2 as c_aacgmv2

    # Test time
    if isinstance(dtime, dt.date):
        if not isinstance(dtime, dt.datetime):
            dtime = dt.datetime.combine(dtime, dt.time(0))
    else:
        raise ValueError('time must be specified as datetime object')

    # Calculate desired location, C routines set date and time
    if m2a:
        # Get the magnetic longitude
        inv_vectorised = np.vectorize(c_aacgmv2.inv_mlt_convert)
        out = inv_vectorised(dtime.year, dtime.month, dtime.day, dtime.hour,
                             dtime.minute, dtime.second, arr)
    else:
        # Get magnetic local time
        mlt_vectorised = np.vectorize(c_aacgmv2.mlt_convert)
        out = mlt_vectorised(dtime.year, dtime.month, dtime.day, dtime.hour,
                             dtime.minute, dtime.second, arr)

    if hasattr(out, "shape") and out.shape == ():
        out = float(out)

    return out



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import, unicode_literals

import datetime as dt
import numpy as np
import pytest
import aacgmv2

class TestPyAACGMV2:

    @classmethod
    def test_module_structure(self):
        """Test module structure"""
        assert aacgmv2
        assert aacgmv2.convert_bool_to_bit
        assert aacgmv2.convert_str_to_bit
        assert aacgmv2.convert_mlt
        assert aacgmv2.convert_latlon
        assert aacgmv2.convert_latlon_arr
        assert aacgmv2.get_aacgm_coord
        assert aacgmv2.get_aacgm_coord_arr
        assert aacgmv2.wrapper
        assert aacgmv2.wrapper.set_coeff_path

    @classmethod
    def test_module_parameters(self):
        """Test module constants"""
        from os import path

        path1 = path.join("aacgmv2", "aacgmv2", "aacgm_coeffs",
                          "aacgm_coeffs-12-")
        if aacgmv2.AACGM_v2_DAT_PREFIX.find(path1) < 0:
            raise AssertionError()

        path2 = path.join("aacgmv2", "aacgmv2", "magmodel_1590-2015.txt")
        if aacgmv2.IGRF_COEFFS.find(path2) < 0:
            raise AssertionError()

        del path1, path2

class TestConvertLatLon:

    def setup(self):
        """Runs before every method to create a clean testing setup"""
        self.dtime = dt.datetime(2015, 1, 1, 0, 0, 0)
        self.ddate = dt.date(2015, 1, 1)
        self.lat_out = None
        self.lon_out = None
        self.r_out = None

    def teardown(self):
        """Runs after every method to clean up previous testing"""
        del self.lat_out, self.lon_out, self.r_out, self.dtime, self.ddate

    def test_convert_latlon(self):
        """Test single value latlon conversion"""
        (self.lat_out, self.lon_out,
         self.r_out) = aacgmv2.convert_latlon(60, 0, 300, self.dtime)
        np.testing.assert_almost_equal(self.lat_out, 58.2258, decimal=4)
        np.testing.assert_almost_equal(self.lon_out, 81.1685, decimal=4)
        np.testing.assert_almost_equal(self.r_out, 1.0457, decimal=4)

    def test_convert_latlon_badidea(self):
        """Test single value latlon conversion with a bad flag"""
        code = "G2A | BADIDEA"
        (self.lat_out, self.lon_out,
         self.r_out) = aacgmv2.convert_latlon(60, 0, 3000, self.dtime, code)
        np.testing.assert_almost_equal(self.lat_out, 64.3568, decimal=4)
        np.testing.assert_almost_equal(self.lon_out, 83.3027, decimal=4)
        np.testing.assert_almost_equal(self.r_out, 1.4694, decimal=4)

        del code

    def test_convert_latlon_location_failure(self):
        """Test single value latlon conversion with a bad location"""
        (self.lat_out, self.lon_out,
         self.r_out) = aacgmv2.convert_latlon(0, 0, 0, self.dtime)
        if not (np.isnan(self.lat_out) & np.isnan(self.lon_out) &
                np.isnan(self.r_out)):
            raise AssertionError()

    def test_convert_latlon_time_failure(self):
        """Test single value latlon conversion with a bad datetime"""
        with pytest.raises(AssertionError):
            (self.lat_out, self.lon_out,
             self.r_out) = aacgmv2.convert_latlon(60, 0, 300, None)

    def test_convert_latlon_datetime_date(self):
        """Test single latlon conversion with date and datetime input"""
        (self.lat_out, self.lon_out,
         self.r_out) = aacgmv2.convert_latlon(60, 0, 300, self.ddate)
        lat_2, lon_2, r_2 = aacgmv2.convert_latlon(60, 0, 300, self.dtime)

        if self.lat_out != lat_2:
            raise AssertionError()
        if self.lon_out != lon_2:
            raise AssertionError()
        if self.r_out != r_2:
            raise AssertionError()

        del lat_2, lon_2, r_2

    def test_warning_below_ground_convert_latlon(self):
        """ Test that a warning is issued if altitude is below zero"""
        import logbook
        lwarn = u"conversion not intended for altitudes < 0 km"

        with logbook.TestHandler() as handler:
            (self.lat_out, self.lon_out,
             self.r_out) = aacgmv2.convert_latlon(60, 0, -1, self.dtime)
            if not handler.has_warning(lwarn):
                raise AssertionError()

        handler.close()

    def test_convert_latlon_maxalt_failure(self):
        """For a single value, test failure for an altitude too high for
        coefficients"""
        (self.lat_out, self.lon_out,
         self.r_out) = aacgmv2.convert_latlon(60, 0, 2001, self.dtime)
        if not (np.isnan(self.lat_out) & np.isnan(self.lon_out) &
                np.isnan(self.r_out)):
            raise AssertionError()

    def test_convert_latlon_lat_failure(self):
        """Test error return for co-latitudes above 90 for a single value"""
        with pytest.raises(AssertionError):
            aacgmv2.convert_latlon(91, 0, 300, self.dtime)

        with pytest.raises(AssertionError):
            aacgmv2.convert_latlon(-91, 0, 300, self.dtime)

class TestConvertLatLonArr:
    def setup(self):
        """Runs before every method to create a clean testing setup"""
        self.dtime = dt.datetime(2015, 1, 1, 0, 0, 0)
        self.ddate = dt.date(2015, 1, 1)
        self.lat_out = None
        self.lon_out = None
        self.r_out = None

    def teardown(self):
        """Runs after every method to clean up previous testing"""
        del self.lat_out, self.lon_out, self.r_out, self.dtime, self.ddate

    def test_convert_latlon_arr_single_val(self):
        """Test array latlon conversion for a single value"""
        (self.lat_out, self.lon_out,
         self.r_out) = aacgmv2.convert_latlon_arr(60, 0, 300, self.dtime)

        if not isinstance(self.lat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.lon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.r_out, np.ndarray):
            raise AssertionError()
        if not (self.r_out.shape == self.lon_out.shape and
                self.lat_out.shape == self.r_out.shape and
                self.r_out.shape == (1,)):
            raise AssertionError()

        np.testing.assert_allclose(self.lat_out, [58.2257709], rtol=1e-4)
        np.testing.assert_allclose(self.lon_out, [81.16846959], rtol=1e-4)
        np.testing.assert_allclose(self.r_out, [1.04566346], rtol=1e-4)

    def test_convert_latlon_arr_list_single(self):
        """Test array latlon conversion for list input of single values"""
        (self.lat_out, self.lon_out,
         self.r_out) = aacgmv2.convert_latlon_arr([60], [0], [300], self.dtime)

        if not isinstance(self.lat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.lon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.r_out, np.ndarray):
            raise AssertionError()
        if not (self.r_out.shape == self.lon_out.shape and
                self.lat_out.shape == self.r_out.shape and
                self.r_out.shape == (1,)):
            raise AssertionError()

        np.testing.assert_allclose(self.lat_out, [58.2257709], rtol=1e-4)
        np.testing.assert_allclose(self.lon_out, [81.16846959], rtol=1e-4)
        np.testing.assert_allclose(self.r_out, [1.04566346], rtol=1e-4)

    def test_convert_latlon_arr_list(self):
        """Test array latlon conversion for list input"""
        (self.lat_out, self.lon_out,
         self.r_out) = aacgmv2.convert_latlon_arr([60, 61], [0, 0], [300, 300],
                                                  self.dtime)

        if not isinstance(self.lat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.lon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.r_out, np.ndarray):
            raise AssertionError()
        if not (self.r_out.shape == self.lon_out.shape and
                self.lat_out.shape == self.r_out.shape and
                self.r_out.shape == (2,)):
            raise AssertionError()

        np.testing.assert_allclose(self.lat_out, [58.22577090, 59.31860933],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.lon_out, [81.16846959, 81.61398933],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.r_out, [1.04566346, 1.04561304],
                                   rtol=1e-4)

    def test_convert_latlon_arr_arr_single(self):
        """Test array latlon conversion for array input of shape (1,)"""
        (self.lat_out, self.lon_out,
        self.r_out) = aacgmv2.convert_latlon_arr(np.array([60]), np.array([0]),
                                                 np.array([300]), self.dtime)

        if not isinstance(self.lat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.lon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.r_out, np.ndarray):
            raise AssertionError()
        if not (self.r_out.shape == self.lon_out.shape and
                self.lat_out.shape == self.r_out.shape and
                self.r_out.shape == (1,)):
            raise AssertionError()

        np.testing.assert_allclose(self.lat_out, [58.2257709], rtol=1e-4)
        np.testing.assert_allclose(self.lon_out, [81.16846959], rtol=1e-4)
        np.testing.assert_allclose(self.r_out, [1.04566346], rtol=1e-4)

    def test_convert_latlon_arr_arr(self):
        """Test array latlon conversion for array input"""
        (self.lat_out, self.lon_out,
         self.r_out) = aacgmv2.convert_latlon_arr(np.array([60, 61]),
                                                  np.array([0, 0]),
                                                  np.array([300, 300]),
                                                  self.dtime)

        if not isinstance(self.lat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.lon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.r_out, np.ndarray):
            raise AssertionError()
        if not (self.r_out.shape == self.lon_out.shape and
                self.lat_out.shape == self.r_out.shape and
                self.r_out.shape == (2,)):
            raise AssertionError()

        np.testing.assert_allclose(self.lat_out, [58.22577090, 59.31860933],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.lon_out, [81.16846959, 81.61398933],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.r_out, [1.04566346, 1.04561304],
                                   rtol=1e-4)

    def test_convert_latlon_arr_list_mix(self):
        """Test array latlon conversion for mixed types with list"""
        (self.lat_out, self.lon_out,
        self.r_out) = aacgmv2.convert_latlon_arr([60, 61], 0, 300, self.dtime)

        if not isinstance(self.lat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.lon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.r_out, np.ndarray):
            raise AssertionError()
        if not (self.r_out.shape == self.lon_out.shape and
                self.lat_out.shape == self.r_out.shape and
                self.r_out.shape == (2,)):
            raise AssertionError()

        np.testing.assert_allclose(self.lat_out, [58.22577090, 59.31860933],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.lon_out, [81.16846959, 81.61398933],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.r_out, [1.04566346, 1.04561304],
                                   rtol=1e-4)

    def test_convert_latlon_arr_arr_mix(self):
        """Test array latlon conversion for mixed type with an array"""
        (self.lat_out, self.lon_out,
        self.r_out) = aacgmv2.convert_latlon_arr(np.array([60, 61]), 0,
                                                 300, self.dtime)

        if not isinstance(self.lat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.lon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.r_out, np.ndarray):
            raise AssertionError()
        if not (self.r_out.shape == self.lon_out.shape and
                self.lat_out.shape == self.r_out.shape and
                self.r_out.shape == (2,)):
            raise AssertionError()

        np.testing.assert_allclose(self.lat_out, [58.22577090, 59.31860933],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.lon_out, [81.16846959, 81.61398933],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.r_out, [1.04566346, 1.04561304],
                                   rtol=1e-4)

    def test_convert_latlon_arr_mult_arr_mix(self):
        """Test array latlon conversion for mix type with multi-dim array"""
        (self.lat_out, self.lon_out,
        self.r_out) = aacgmv2.convert_latlon_arr(np.array([[60, 61, 62],
                                                           [63, 64, 65]]),
                                                 0, 300, self.dtime)

        if not isinstance(self.lat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.lon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.r_out, np.ndarray):
            raise AssertionError()
        if not (self.r_out.shape == self.lon_out.shape and
                self.lat_out.shape == self.r_out.shape and
                self.r_out.shape == (2, 3)):
            raise AssertionError()

        np.testing.assert_allclose(self.lat_out,
                                   [[58.2257709, 59.3186093, 60.4039740],
                                    [61.4819893, 62.5527635, 63.6163840]],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.lon_out,
                                   [[81.1684696, 81.6139893, 82.0871880],
                                    [82.5909499, 83.1285895, 83.7039272]],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.r_out,
                                   [[1.04566346, 1.04561304, 1.04556369],
                                    [1.04551548, 1.04546847, 1.04542272]],
                                   rtol=1e-4)

    def test_convert_latlon_arr_mult_arr_unequal(self):
        """Test array latlon conversion for unequal sized multi-dim array"""
        (self.lat_out, self.lon_out,
         self.r_out) = aacgmv2.convert_latlon_arr(np.array([[60, 61, 62],
                                                            [63, 64, 65]]),
                                                  np.array([0]),
                                                  np.array([300]), self.dtime)

        if not isinstance(self.lat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.lon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.r_out, np.ndarray):
            raise AssertionError()
        if not (self.r_out.shape == self.lon_out.shape and
                self.lat_out.shape == self.r_out.shape and
                self.r_out.shape == (2, 3)):
            raise AssertionError()

        np.testing.assert_allclose(self.lat_out,
                                   [[58.2257709, 59.3186093, 60.4039740],
                                    [61.4819893, 62.5527635, 63.6163840]],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.lon_out,
                                   [[81.1684696, 81.6139893, 82.0871880],
                                    [82.5909499, 83.1285895, 83.7039272]],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.r_out,
                                   [[1.04566346, 1.04561304, 1.04556369],
                                    [1.04551548, 1.04546847, 1.04542272]],
                                   rtol=1e-4)

    def test_convert_latlon_arr_badidea(self):
        """Test array latlon conversion for BADIDEA"""
        code = "G2A | BADIDEA"
        (self.lat_out, self.lon_out,
         self.r_out) = aacgmv2.convert_latlon_arr([60], [0], [3000],
                                                  self.dtime, code)

        np.testing.assert_allclose(self.lat_out, [64.35677791], rtol=1e-4)
        np.testing.assert_allclose(self.lon_out, [83.30272053], rtol=1e-4)
        np.testing.assert_allclose(self.r_out, [1.46944431], rtol=1e-4)

    def test_convert_latlon_arr_location_failure(self):
        """Test array latlon conversion with a bad location"""
        (self.lat_out, self.lon_out,
         self.r_out) = aacgmv2.convert_latlon_arr([0], [0], [0], self.dtime)

        if not isinstance(self.lat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.lon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.r_out, np.ndarray):
            raise AssertionError()
        if not (self.r_out.shape == self.lon_out.shape and
                self.lat_out.shape == self.r_out.shape and
                self.r_out.shape == (1,)):
            raise AssertionError()
        if not np.all([np.isnan(self.lat_out), np.isnan(self.lon_out),
                       np.isnan(self.r_out)]):
            raise AssertionError()

    def test_convert_latlon_arr_time_failure(self):
        """Test array latlon conversion with a bad time"""
        with pytest.raises(AssertionError):
            (self.lat_out, self.lon_out,
             self.r_out) = aacgmv2.convert_latlon_arr([60], [0], [300], None)

    def test_convert_latlon_arr_datetime_date(self):
        """Test array latlon conversion with date and datetime input"""
        (self.lat_out, self.lon_out,
         self.r_out) = aacgmv2.convert_latlon_arr([60], [0], [300], self.ddate)
        lat_2, lon_2, r_2 = aacgmv2.convert_latlon_arr([60], [0], [300],
                                                       self.dtime)
        if self.lat_out != lat_2:
            raise AssertionError()
        if self.lon_out != lon_2:
            raise AssertionError()
        if self.r_out != r_2:
            raise AssertionError()

        del lat_2, lon_2, r_2

    def test_warning_below_ground_convert_latlon_arr(self):
        """ Test that a warning is issued if altitude is below zero"""
        import logbook
        lwarn = u"conversion not intended for altitudes < 0 km"

        with logbook.TestHandler() as handler:
            (self.lat_out, self.lon_out,
             self.r_out) = aacgmv2.convert_latlon_arr([60], [0], [-1],
                                                      self.dtime)
            if not handler.has_warning(lwarn):
                raise AssertionError()

        handler.close()

    def test_convert_latlon_arr_maxalt_failure(self):
        """For an array, test failure for an altitude too high for
        coefficients"""
        (self.lat_out, self.lon_out,
         self.r_out) = aacgmv2.convert_latlon_arr([60], [0], [2001], self.dtime)
        if not np.all([np.isnan(self.lat_out), np.isnan(self.lon_out),
                       np.isnan(self.r_out)]):
            raise AssertionError()

    def test_convert_latlon_arr_lat_failure(self):
        """Test error return for co-latitudes above 90 for an array"""
        with pytest.raises(AssertionError):
            aacgmv2.convert_latlon_arr([91, 60, -91], 0, 300, self.dtime)

class TestGetAACGMCoord:
    def setup(self):
        """Runs before every method to create a clean testing setup"""
        self.dtime = dt.datetime(2015, 1, 1, 0, 0, 0)
        self.ddate = dt.date(2015, 1, 1)
        self.mlat_out = None
        self.mlon_out = None
        self.mlt_out = None

    def teardown(self):
        """Runs after every method to clean up previous testing"""
        del self.mlat_out, self.mlon_out, self.mlt_out, self.dtime, self.ddate

    def test_get_aacgm_coord(self):
        """Test single value AACGMV2 calculation, defaults to TRACE"""
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord(60, 0, 300, self.dtime)

        np.testing.assert_almost_equal(self.mlat_out, 58.2247, decimal=4)
        np.testing.assert_almost_equal(self.mlon_out, 81.1761, decimal=4)
        np.testing.assert_almost_equal(self.mlt_out, 0.1889, decimal=4)

    def test_get_aacgm_coord_badidea(self):
        """Test single value AACGMV2 calculation with a bad flag"""
        method = "BADIDEA"
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord(60, 0, 3000, self.dtime,
                                                 method=method)

        np.testing.assert_almost_equal(self.mlat_out, 64.3568, decimal=4)
        np.testing.assert_almost_equal(self.mlon_out, 83.3027, decimal=4)
        np.testing.assert_almost_equal(self.mlt_out, 0.3307, decimal=4)
        del method

    def test_get_aacgm_coord_location_failure(self):
        """Test single value AACGMV2 calculation with a bad location"""
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord(0, 0, 0, self.dtime)
        if not (np.isnan(self.mlat_out) & np.isnan(self.mlon_out) &
                np.isnan(self.mlt_out)):
            raise AssertionError()

    def test_get_aacgm_coord_time_failure(self):
        """Test single value AACGMV2 calculation with a bad datetime"""
        with pytest.raises(AssertionError):
            (self.mlat_out, self.mlon_out,
             self.mlt_out) = aacgmv2.get_aacgm_coord(60, 0, 300, None)

    def test_get_aacgm_coord_datetime_date(self):
        """Test single AACGMV2 calculation with date and datetime input"""
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord(60, 0, 300, self.ddate)
        mlat_2, mlon_2, mlt_2 = aacgmv2.get_aacgm_coord(60, 0, 300, self.dtime)

        np.testing.assert_almost_equal(self.mlat_out, mlat_2, decimal=6)
        np.testing.assert_almost_equal(self.mlon_out, mlon_2, decimal=6)
        np.testing.assert_almost_equal(self.mlt_out, mlt_2, decimal=6)

        del mlat_2, mlon_2, mlt_2

    def test_warning_below_ground_get_aacgm_coord(self):
        """ Test that a warning is issued if altitude is below zero"""
        import logbook
        lwarn = u"conversion not intended for altitudes < 0 km"

        with logbook.TestHandler() as handler:
            (self.mlat_out, self.mlon_out,
             self.mlt_out) = aacgmv2.get_aacgm_coord(60, 0, -1, self.dtime)
            if not handler.has_warning(lwarn):
                raise AssertionError()

        handler.close()

    def test_get_aacgm_coord_maxalt_failure(self):
        """For a single value, test failure for an altitude too high for
        coefficients"""
        method = ""
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord(60, 0, 2001, self.dtime,
                                                 method=method)
        if not (np.isnan(self.mlat_out) & np.isnan(self.mlon_out) &
                np.isnan(self.mlt_out)):
            raise AssertionError()

    def test_get_aacgm_coord_mlat_failure(self):
        """Test error return for co-latitudes above 90 for a single value"""
        import logbook
        lerr = u"unrealistic latitude"

        with logbook.TestHandler() as hhigh:
            with pytest.raises(AssertionError):
                (self.mlat_out, self.mlon_out,
                 self.mlt_out) = aacgmv2.get_aacgm_coord(91, 0, 300, self.dtime)
                if not hhigh.has_error(lerr):
                    raise AssertionError()

        with logbook.TestHandler() as hlow:
            with pytest.raises(AssertionError):
                (self.mlat_out, self.mlon_out,
                 self.mlt_out) = aacgmv2.get_aacgm_coord(-91, 0, 300,
                                                         self.dtime)
                if not hlow.has_error(lerr):
                    raise AssertionError()

        hhigh.close()
        hlow.close()

class TestGetAACGMCoordArr:
    def setup(self):
        """Runs before every method to create a clean testing setup"""
        self.dtime = dt.datetime(2015, 1, 1, 0, 0, 0)
        self.ddate = dt.date(2015, 1, 1)
        self.mlat_out = None
        self.mlon_out = None
        self.mlt_out = None

    def teardown(self):
        """Runs after every method to clean up previous testing"""
        del self.mlat_out, self.mlon_out, self.mlt_out, self.dtime, self.ddate

    def test_get_aacgm_coord_arr_single_val(self):
        """Test array AACGMV2 calculation for a single value"""
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord_arr(60, 0, 300, self.dtime)

        if not isinstance(self.mlat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlt_out, np.ndarray):
            raise AssertionError()
        if not (self.mlt_out.shape == self.mlon_out.shape and
                self.mlat_out.shape == self.mlt_out.shape and
                self.mlt_out.shape == (1,)):
            raise AssertionError()

        np.testing.assert_allclose(self.mlat_out, [58.22474610], rtol=1e-4)
        np.testing.assert_allclose(self.mlon_out, [81.17611033], rtol=1e-4)
        np.testing.assert_allclose(self.mlt_out, [0.18891995], rtol=1e-4)

    def test_get_aacgm_coord_arr_list_single(self):
        """Test array AACGMV2 calculation for list input of single values"""
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord_arr([60], [0], [300],
                                                     self.dtime)

        if not isinstance(self.mlat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlt_out, np.ndarray):
            raise AssertionError()
        if not (self.mlt_out.shape == self.mlon_out.shape and
                self.mlat_out.shape == self.mlt_out.shape and
                self.mlt_out.shape == (1,)):
            raise AssertionError()

        np.testing.assert_allclose(self.mlat_out, [58.22474610], rtol=1e-4)
        np.testing.assert_allclose(self.mlon_out, [81.17611033], rtol=1e-4)
        np.testing.assert_allclose(self.mlt_out, [0.18891995], rtol=1e-4)

    def test_get_aacgm_coord_arr_list(self):
        """Test array AACGMV2 calculation for list input"""
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord_arr([60, 61], [0, 0],
                                                     [300, 300], self.dtime)

        if not isinstance(self.mlat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlt_out, np.ndarray):
            raise AssertionError()
        if not (self.mlt_out.shape == self.mlon_out.shape and
                self.mlat_out.shape == self.mlt_out.shape and
                self.mlt_out.shape == (2,)):
            raise AssertionError()

        np.testing.assert_allclose(self.mlat_out,
                                   [58.22474610, 59.31648007], rtol=1e-4)
        np.testing.assert_allclose(self.mlon_out,
                                   [81.17611033, 81.62281360], rtol=1e-4)
        np.testing.assert_allclose(self.mlt_out,
                                   [0.18891995, 0.21870017], rtol=1e-4)

    def test_get_aacgm_coord_arr_arr_single(self):
        """Test array AACGMV2 calculation for array with a single value"""
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord_arr(np.array([60]),
                                                     np.array([0]),
                                                     np.array([300]),
                                                     self.dtime)

        if not isinstance(self.mlat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlt_out, np.ndarray):
            raise AssertionError()
        if not (self.mlt_out.shape == self.mlon_out.shape and
                self.mlat_out.shape == self.mlt_out.shape and
                self.mlt_out.shape == (1,)):
            raise AssertionError()

        np.testing.assert_almost_equal(self.mlat_out, 58.2247, decimal=4)
        np.testing.assert_almost_equal(self.mlon_out, 81.1761, decimal=4)
        np.testing.assert_almost_equal(self.mlt_out, 0.1889, decimal=4)

    def test_get_aacgm_coord_arr_arr(self):
        """Test array AACGMV2 calculation for an array"""
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord_arr(np.array([60, 61]),
                                                     np.array([0, 0]),
                                                     np.array([300, 300]),
                                                     self.dtime)

        if not isinstance(self.mlat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlt_out, np.ndarray):
            raise AssertionError()
        if not (self.mlt_out.shape == self.mlon_out.shape and
                self.mlat_out.shape == self.mlt_out.shape and
                self.mlt_out.shape == (2,)):
            raise AssertionError()

        np.testing.assert_allclose(self.mlat_out,
                                   [58.22474610, 59.31648007], rtol=1e-4)
        np.testing.assert_allclose(self.mlon_out,
                                   [81.17611033, 81.62281360], rtol=1e-4)
        np.testing.assert_allclose(self.mlt_out,
                                   [0.18891995, 0.21870017], rtol=1e-4)

    def test_get_aacgm_coord_arr_list_mix(self):
        """Test array AACGMV2 calculation for a list and floats"""
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord_arr([60, 61], 0, 300,
                                                     self.dtime)

        if not isinstance(self.mlat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlt_out, np.ndarray):
            raise AssertionError()
        if not (self.mlt_out.shape == self.mlon_out.shape and
                self.mlat_out.shape == self.mlt_out.shape and
                self.mlt_out.shape == (2,)):
            raise AssertionError()

        np.testing.assert_allclose(self.mlat_out,
                                   [58.22474610, 59.31648007], rtol=1e-4)
        np.testing.assert_allclose(self.mlon_out,
                                   [81.17611033, 81.62281360], rtol=1e-4)
        np.testing.assert_allclose(self.mlt_out,
                                   [0.18891995, 0.21870017], rtol=1e-4)

    def test_get_aacgm_coord_arr_arr_mix(self):
        """Test array AACGMV2 calculation for an array and floats"""
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord_arr(np.array([60, 61]), 0,
                                                     300, self.dtime)

        if not isinstance(self.mlat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlt_out, np.ndarray):
            raise AssertionError()
        if not (self.mlt_out.shape == self.mlon_out.shape and
                self.mlat_out.shape == self.mlt_out.shape and
                self.mlt_out.shape == (2,)):
            raise AssertionError()

        np.testing.assert_allclose(self.mlat_out,
                                   [58.22474610, 59.31648007], rtol=1e-4)
        np.testing.assert_allclose(self.mlon_out,
                                   [81.17611033, 81.62281360], rtol=1e-4)
        np.testing.assert_allclose(self.mlt_out,
                                   [0.18891995, 0.21870017], rtol=1e-4)

    def test_get_aacgm_coord_arr_mult_arr_mix(self):
        """Test array AACGMV2 calculation for a multi-dim array and
        floats"""
        mlat_in = np.array([[60, 61, 62], [63, 64, 65]])
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord_arr(mlat_in, 0, 300,
                                                     self.dtime)

        if not isinstance(self.mlat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlt_out, np.ndarray):
            raise AssertionError()
        if not (self.mlt_out.shape == self.mlon_out.shape and
                self.mlat_out.shape == self.mlt_out.shape and
                self.mlt_out.shape == (2, 3)):
            raise AssertionError()

        np.testing.assert_allclose(self.mlat_out,
                                   [[58.2247461, 59.3164801, 60.4008651],
                                    [61.4780560, 62.5481858, 63.6113609]],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.mlon_out,
                                   [[81.1761103, 81.6228136, 82.0969646],
                                    [82.6013918, 83.1393547, 83.7146224]],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.mlt_out,
                                   [[0.18891995, 0.21870017, 0.25031024],
                                    [0.28393872, 0.31980291, 0.35815409]],
                                   rtol=1e-4)
        del mlat_in

    def test_get_aacgm_coord_arr_arr_unequal(self):
        """Test array AACGMV2 calculation for unequal arrays"""
        mlat_in = np.array([[60, 61, 62], [63, 64, 65]])
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord_arr(mlat_in, np.array([0]),
                                                     np.array([300]),
                                                     self.dtime)

        if not isinstance(self.mlat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlt_out, np.ndarray):
            raise AssertionError()
        if not (self.mlt_out.shape == self.mlon_out.shape and
                self.mlat_out.shape == self.mlt_out.shape and
                self.mlt_out.shape == (2, 3)):
            raise AssertionError()

        np.testing.assert_allclose(self.mlat_out,
                                   [[58.2247, 59.3165, 60.4009],
                                    [61.4781, 62.5482, 63.6114]], rtol=1e-3)
        np.testing.assert_allclose(self.mlon_out,
                                   [[81.1761, 81.6228, 82.0970],
                                    [82.6014, 83.1394, 83.7146]], rtol=1e-3)
        np.testing.assert_allclose(self.mlt_out,
                                   [[0.1889, 0.2187, 0.2503],
                                    [0.2839, 0.3198, 0.3582]], rtol=1e-3)
        del mlat_in

    def test_get_aacgm_coord_arr_badidea(self):
        """Test array AACGMV2 calculation for BADIDEA"""
        method = "BADIDEA"
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord_arr([60], [0], [3000],
                                                     self.dtime, method=method)

        np.testing.assert_allclose(self.mlat_out, [64.35677791], rtol=1e-3)
        np.testing.assert_allclose(self.mlon_out, [83.30272053], rtol=1e-3)
        np.testing.assert_allclose(self.mlt_out, [0.33069397], rtol=1e-3)

        del method

    def test_get_aacgm_coord_arr_location_failure(self):
        """Test array AACGMV2 calculation with a bad location"""
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord_arr([0], [0], [0], self.dtime)

        if not isinstance(self.mlat_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlon_out, np.ndarray):
            raise AssertionError()
        if not isinstance(self.mlt_out, np.ndarray):
            raise AssertionError()
        if not (self.mlt_out.shape == self.mlon_out.shape and
                self.mlat_out.shape == self.mlt_out.shape and
                self.mlt_out.shape == (1,)):
            raise AssertionError()
        if not np.all([np.isnan(self.mlat_out), np.isnan(self.mlon_out),
                       np.isnan(self.mlt_out)]):
            raise AssertionError()

    def test_get_aacgm_coord_arr_time_failure(self):
        """Test array AACGMV2 calculation with a bad time"""
        with pytest.raises(AssertionError):
            (self.mlat_out, self.mlon_out,
             self.mlt_out) = aacgmv2.get_aacgm_coord_arr([60], [0], [300],
                                                         None)

    def test_get_aacgm_coord_arr_datetime_date(self):
        """Test array AACGMV2 calculation with date and datetime input"""
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord_arr([60], [0], [300],
                                                     self.ddate)
        mlat_2, mlon_2, mlt_2 = aacgmv2.get_aacgm_coord_arr([60], [0], [300],
                                                            self.dtime)

        np.testing.assert_almost_equal(self.mlat_out, mlat_2, decimal=6)
        np.testing.assert_almost_equal(self.mlon_out, mlon_2, decimal=6)
        np.testing.assert_almost_equal(self.mlt_out, mlt_2, decimal=6)

        del mlat_2, mlon_2, mlt_2

    def test_warning_below_ground_get_aacgm_coord_arr(self):
        """ Test that a warning is issued if altitude is below zero"""
        import logbook
        lwarn = u"conversion not intended for altitudes < 0 km"

        with logbook.TestHandler() as handler:
            (self.mlat_out, self.mlon_out,
             self.mlt_out) = aacgmv2.get_aacgm_coord_arr([60], [0], [-1],
                                                         self.dtime)
            if not handler.has_warning(lwarn):
                raise AssertionError()

        handler.close()

    def test_get_aacgm_coord_arr_maxalt_failure(self):
        """For an array, test failure for an altitude too high for
        coefficients"""
        method = ""
        (self.mlat_out, self.mlon_out,
         self.mlt_out) = aacgmv2.get_aacgm_coord_arr([60], [0], [2001],
                                                     self.dtime, method=method)
        if not np.all([np.isnan(self.mlat_out), np.isnan(self.mlon_out),
                       np.isnan(self.mlt_out)]):
            raise AssertionError()

        del method

    def test_get_aacgm_coord_arr_mlat_failure(self):
        """Test error return for co-latitudes above 90 for an array"""
        import logbook
        lerr = u"unrealistic latitude"

        with logbook.TestHandler() as handler:
            with pytest.raises(AssertionError):
                aacgmv2.get_aacgm_coord_arr([91, 60, -91], 0, 300,
                                                self.dtime)
                if not handler.has_error(lerr):
                    raise AssertionError()

        handler.close()

class TestConvertCode:
    @classmethod
    def test_convert_str_to_bit_g2a(self):
        """Test conversion from string code to bit G2A"""
        if aacgmv2.convert_str_to_bit("G2A") != aacgmv2._aacgmv2.G2A:
            raise AssertionError()

    @classmethod
    def test_convert_str_to_bit_a2g(self):
        """Test conversion from string code to bit A2G"""
        if aacgmv2.convert_str_to_bit("A2G") != aacgmv2._aacgmv2.A2G:
            raise AssertionError()

    @classmethod
    def test_convert_str_to_bit_trace(self):
        """Test conversion from string code to bit TRACE"""
        if aacgmv2.convert_str_to_bit("TRACE") != aacgmv2._aacgmv2.TRACE:
            raise AssertionError()

    @classmethod
    def test_convert_str_to_bit_allowtrace(self):
        """Test conversion from string code to bit ALLOWTRACE"""
        if(aacgmv2.convert_str_to_bit("ALLOWTRACE") !=
           aacgmv2._aacgmv2.ALLOWTRACE):
            raise AssertionError()

    @classmethod
    def test_convert_str_to_bit_badidea(self):
        """Test conversion from string code to bit BADIDEA"""
        if(aacgmv2.convert_str_to_bit("BADIDEA") !=
           aacgmv2._aacgmv2.BADIDEA):
            raise AssertionError()

    @classmethod
    def test_convert_str_to_bit_geocentric(self):
        """Test conversion from string code to bit GEOCENTRIC"""
        if(aacgmv2.convert_str_to_bit("GEOCENTRIC") !=
           aacgmv2._aacgmv2.GEOCENTRIC):
            raise AssertionError()

    @classmethod
    def test_convert_str_to_bit_lowercase(self):
        """Test conversion from string code to bit for a lowercase code"""
        if aacgmv2.convert_str_to_bit("g2a") != aacgmv2._aacgmv2.G2A:
            raise AssertionError()

    @classmethod
    def test_convert_str_to_bit_spaces(self):
        """Test conversion from string code to bit for a code with spaces"""
        if(aacgmv2.convert_str_to_bit("G2A | trace") !=
           aacgmv2._aacgmv2.G2A + aacgmv2._aacgmv2.TRACE):
            raise AssertionError()

    @classmethod
    def test_convert_str_to_bit_invalid(self):
        """Test conversion from string code to bit for an invalid code"""
        if aacgmv2.convert_str_to_bit("ggoogg|") != aacgmv2._aacgmv2.G2A:
            raise AssertionError()

    @classmethod
    def test_convert_bool_to_bit_g2a(self):
        """Test conversion from string code to bit G2A"""
        if aacgmv2.convert_bool_to_bit() != aacgmv2._aacgmv2.G2A:
            raise AssertionError()

    @classmethod
    def test_convert_bool_to_bit_a2g(self):
        """Test conversion from string code to bit A2G"""
        if aacgmv2.convert_bool_to_bit(a2g=True) != aacgmv2._aacgmv2.A2G:
            raise AssertionError()

    @classmethod
    def test_convert_bool_to_bit_trace(self):
        """Test conversion from string code to bit TRACE"""
        if aacgmv2.convert_bool_to_bit(trace=True) != aacgmv2._aacgmv2.TRACE:
            raise AssertionError()

    @classmethod
    def test_convert_bool_to_bit_allowtrace(self):
        """Test conversion from string code to bit ALLOWTRACE"""
        if(aacgmv2.convert_bool_to_bit(allowtrace=True) !=
           aacgmv2._aacgmv2.ALLOWTRACE):
            raise AssertionError()

    @classmethod
    def test_convert_bool_to_bit_badidea(self):
        """Test conversion from string code to bit BADIDEA"""
        if(aacgmv2.convert_bool_to_bit(badidea=True) !=
           aacgmv2._aacgmv2.BADIDEA):
            raise AssertionError()

    @classmethod
    def test_convert_bool_to_bit_geocentric(self):
        """Test conversion from string code to bit GEOCENTRIC"""
        if(aacgmv2.convert_bool_to_bit(geocentric=True) !=
           aacgmv2._aacgmv2.GEOCENTRIC):
            raise AssertionError()

class TestMLTConvert:
    def setup(self):
        """Runs before every method to create a clean testing setup"""
        self.dtime = dt.datetime(2015, 1, 1, 0, 0, 0)
        self.dtime2 = dt.datetime(2015, 1, 1, 10, 0, 0)
        self.ddate = dt.date(2015, 1, 1)
        self.mlon_out = None
        self.mlt_out = None
        self.mlt_diff = None
        self.mlon_list = [270.0, 80.0, -95.0]
        self.mlt_list = [12.0, 25.0, -1.0]
        self.mlon_comp = [-101.657689, 93.34231102, 63.34231102]
        self.mlt_comp = [12.77717927, 0.1105126, 12.44384593]
        self.diff_comp = np.ones(shape=(3,)) * -10.52411552

    def teardown(self):
        """Runs after every method to clean up previous testing"""
        del self.mlon_out, self.mlt_out, self.mlt_list, self.mlon_list
        del self.mlon_comp, self.mlt_comp, self.mlt_diff, self.diff_comp

    def test_date_input(self):
        """Test to see that the date input works"""
        self.mlt_out = aacgmv2.convert_mlt(self.mlon_list, self.ddate,
                                           m2a=False)
        np.testing.assert_allclose(self.mlt_out, self.mlt_comp, rtol=1.0e-4)

    def test_datetime_exception(self):
        """Test to see that a value error is raised with bad time input"""
        with pytest.raises(ValueError):
            self.mlt_out = aacgmv2.wrapper.convert_mlt(self.mlon_list, 1997)

    def test_inv_convert_mlt_single(self):
        """Test MLT inversion for a single value"""
        for i,mlt in enumerate(self.mlt_list):
            self.mlon_out = aacgmv2.convert_mlt(mlt, self.dtime, m2a=True)
            np.testing.assert_almost_equal(self.mlon_out, self.mlon_comp[i],
                                           decimal=4)

    def test_inv_convert_mlt_list(self):
        """Test MLT inversion for a list"""
        self.mlon_out = aacgmv2.convert_mlt(self.mlt_list, self.dtime, m2a=True)
        np.testing.assert_allclose(self.mlon_out, self.mlon_comp, rtol=1.0e-4)

    def test_inv_convert_mlt_arr(self):
        """Test MLT inversion for an array"""
        self.mlon_out = aacgmv2.convert_mlt(np.array(self.mlt_list), self.dtime,
                                            m2a=True)

        np.testing.assert_allclose(self.mlon_out, self.mlon_comp, rtol=1.0e-4)

    def test_inv_convert_mlt_wrapping(self):
        """Test MLT wrapping"""
        self.mlon_out = aacgmv2.convert_mlt(np.array([1, 25, -1, 23]),
                                            self.dtime, m2a=True)

        np.testing.assert_almost_equal(self.mlon_out[0], self.mlon_out[1],
                                       decimal=6)
        np.testing.assert_almost_equal(self.mlon_out[2], self.mlon_out[3],
                                       decimal=6)

    def test_mlt_convert_mlon_wrapping(self):
        """Test mlon wrapping"""
        self.mlt_out = aacgmv2.convert_mlt(np.array([270, -90, 1, 361]),
                                           self.dtime, m2a=False)

        np.testing.assert_almost_equal(self.mlt_out[0], self.mlt_out[1],
                                       decimal=6)
        np.testing.assert_almost_equal(self.mlt_out[2], self.mlt_out[3],
                                       decimal=6)

    def test_mlt_convert_single(self):
        """Test MLT calculation for a single value"""
        for i,mlon in enumerate(self.mlon_list):
            self.mlt_out = aacgmv2.convert_mlt(mlon, self.dtime, m2a=False)
            np.testing.assert_almost_equal(self.mlt_out, self.mlt_comp[i],
                                           decimal=4)

    def test_mlt_convert_list(self):
        """Test MLT calculation for a list"""
        self.mlt_out = aacgmv2.convert_mlt(self.mlon_list, self.dtime,
                                           m2a=False)
        np.testing.assert_allclose(self.mlt_out, self.mlt_comp, rtol=1.0e-4)

    def test_mlt_convert_arr(self):
        """Test MLT calculation for an array"""
        self.mlt_out = aacgmv2.convert_mlt(np.array(self.mlon_list),
                                           self.dtime, m2a=False)
        np.testing.assert_allclose(self.mlt_out, self.mlt_comp, rtol=1.0e-4)

    def test_mlt_convert_change(self):
        """Test that MLT changes with UT"""
        self.mlt_out = aacgmv2.convert_mlt(self.mlon_list, self.dtime)
        self.mlt_diff = self.mlt_out - aacgmv2.convert_mlt(self.mlon_list,
                                                           self.dtime2)

        np.testing.assert_allclose(self.mlt_diff, self.diff_comp, rtol=1.0e-4)

class TestCoeffPath:

    def setup(self):
        """Runs before every method to create a clean testing setup"""
        import os
        os.environ['IGRF_COEFFS'] = "default_igrf"
        os.environ['AACGM_v2_DAT_PREFIX'] = "default_coeff"
        self.default_igrf = os.environ['IGRF_COEFFS']
        self.default_coeff = os.environ['AACGM_v2_DAT_PREFIX']

    def teardown(self):
        """Runs after every method to clean up previous testing"""
        del self.default_igrf, self.default_coeff

    def test_set_coeff_path_default(self):
        """Test the coefficient path setting using default values"""
        import os
        aacgmv2.wrapper.set_coeff_path()

        if os.environ['IGRF_COEFFS'] != self.default_igrf:
            raise AssertionError()
        if os.environ['AACGM_v2_DAT_PREFIX'] != self.default_coeff:
            raise AssertionError()

    @classmethod
    def test_set_coeff_path_string(self):
        """Test the coefficient path setting using two user specified values"""
        import os
        aacgmv2.wrapper.set_coeff_path("hi", "bye")

        if os.environ['IGRF_COEFFS'] != "hi":
            raise AssertionError()
        if os.environ['AACGM_v2_DAT_PREFIX'] != "bye":
            raise AssertionError()

    @classmethod
    def test_set_coeff_path_true(self):
        """Test the coefficient path setting using the module values"""
        import os
        aacgmv2.wrapper.set_coeff_path(True, True)

        if os.environ['IGRF_COEFFS'] != aacgmv2.IGRF_COEFFS:
            raise AssertionError()
        if os.environ['AACGM_v2_DAT_PREFIX'] != aacgmv2.AACGM_v2_DAT_PREFIX:
            raise AssertionError()

    def test_set_only_aacgm_coeff_path(self):
        """Test the coefficient path setting using a mix of input"""
        import os
        aacgmv2.wrapper.set_coeff_path(coeff_prefix="hi")

        if os.environ['IGRF_COEFFS'] != self.default_igrf:
            raise AssertionError()
        if os.environ['AACGM_v2_DAT_PREFIX'] != "hi":
            raise AssertionError()

    def test_set_only_igrf_coeff_path(self):
        """Test the coefficient path setting using a mix of input"""
        import os
        aacgmv2.wrapper.set_coeff_path(igrf_file="hi")

        if os.environ['IGRF_COEFFS'] != "hi":
            raise AssertionError()
        if os.environ['AACGM_v2_DAT_PREFIX'] != self.default_coeff:
            raise AssertionError()

    @classmethod
    def test_set_both_mixed(self):
        """Test the coefficient path setting using a mix of input"""
        import os
        aacgmv2.wrapper.set_coeff_path(igrf_file=True, coeff_prefix="hi")

        if os.environ['IGRF_COEFFS'] != aacgmv2.IGRF_COEFFS:
            raise AssertionError()
        if os.environ['AACGM_v2_DAT_PREFIX'] != "hi":
            raise AssertionError()



# -*- coding: utf-8 -*-
"""aacgmv2

Modules
---------------------------------------------------------------------------
_aacgmv2 : Contains functions and variables from c code
deprecated : Contains deprecated functions from previous versions
wrapper : Contains current python functions
---------------------------------------------------------------------------

Parameters
---------------------------------------------------------------------------
AACGM_V2_DAT_PREFIX
IGRF_12_COEFFS
_aacgmv2.G2A
_aacgmv2.A2G
_aacgmv2.TRACE
_aacgmv2.ALLOWTRACE
_aacgmv2.BADIDEA
_aacgmv2.GEOCENTRIC
---------------------------------------------------------------------------

Functions
---------------------------------------------------------------------------
convert_latlon_arr
convert_str_to_bit
convert_bool_to_bit
get_aacgm_coord
get_aacgm_coord_arr
convert
convert_mlt
wrapper.set_coeff_path
deprecated.subsol
_aacgmv2.convert
_aacgmv2.set_datetime
_aacgmv2.mlt_convert
_aacgmv2.mlt_convert_yrsec
_aacgmv2.inv_mlt_convert
_aacgmv2.inv_mlt_convert_yrsec
---------------------------------------------------------------------------
"""
from __future__ import print_function
import os as _os
import logbook as logging

__version__ = "2.5.0"

# path and filename prefix for the IGRF coefficients
AACGM_v2_DAT_PREFIX = _os.path.join(_os.path.realpath( \
                                                _os.path.dirname(__file__)),
                                    'aacgm_coeffs', 'aacgm_coeffs-12-')
IGRF_COEFFS = _os.path.join(_os.path.realpath(_os.path.dirname(__file__)),
                            'magmodel_1590-2015.txt')

# If not defined, set the IGRF and AACGM environment variables
reset_warn = False
if 'IGRF_COEFFS' in _os.environ.keys():
    print("resetting environment variable IGRF_COEFFS in python script")
    reset_warn = True
_os.environ['IGRF_COEFFS'] = IGRF_COEFFS

if 'AACGM_v2_DAT_PREFIX' in _os.environ.keys():
    print("resetting environment variable AACGM_v2_DAT_PREFIX in python script")
    reset_warn = True
_os.environ['AACGM_v2_DAT_PREFIX'] = AACGM_v2_DAT_PREFIX

if reset_warn:
    print("non-default coefficient files may be specified by running " +
          "aacgmv2.wrapper.set_coeff_path before any other functions")
# Imports
#---------------------------------------------------------------------

try:
    from aacgmv2.wrapper import (convert_latlon, convert_mlt, get_aacgm_coord)
    from aacgmv2.wrapper import (convert_latlon_arr, get_aacgm_coord_arr)
    from aacgmv2.wrapper import (convert_bool_to_bit, convert_str_to_bit)
except Exception as err:
    logging.exception(__file__ + ' -> aacgmv2: ' + str(err))

try:
    from aacgmv2 import (deprecated)
    from aacgmv2.deprecated import (convert)
except Exception as err:
    logging.exception(__file__ + ' -> aacgmv2: ' + str(err))

try:
    from aacgmv2 import (_aacgmv2)
except Exception as err:
    logging.exception(__file__ + ' -> aacgmv2: ' + str(err))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import os
import sys

if __name__ == "__main__":
    base_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    print("Project path: {0}".format(base_path))
    env_path = os.path.join(base_path, ".tox", "bootstrap")
    if sys.platform == "win32":
        bin_path = os.path.join(env_path, "Scripts")
    else:
        bin_path = os.path.join(env_path, "bin")
    if not os.path.exists(env_path):
        import subprocess
        print("Making bootstrap env in: {0} ...".format(env_path))
        try:
            subprocess.check_call(["virtualenv", env_path])
        except Exception:
            subprocess.check_call([sys.executable, "-m", "virtualenv",
                                   env_path])
        print("Installing `jinja2` and `matrix` into bootstrap environment ...")
        subprocess.check_call([os.path.join(bin_path, "pip"), "install",
                               "jinja2", "matrix"])
    activate = os.path.join(bin_path, "activate_this.py")
    exec(compile(open(activate, "rb").read(), activate, "exec"),
         dict(__file__=activate))

    import jinja2
    import matrix

    jinja = jinja2.Environment(
        loader=jinja2.FileSystemLoader(os.path.join(base_path, "ci",
                                                    "templates")),
        trim_blocks=True,
        lstrip_blocks=True,
        keep_trailing_newline=True
    )
    tox_envs = {}
    for (alias, conf) in matrix.from_file(os.path.join(base_path,
                                                       "setup.cfg")).items():
        python = conf["python_versions"]
        #deps = conf["dependencies"]
        if "coverage_flags" in conf:
            cover = {"false": False,
                     "true": True}[conf["coverage_flags"].lower()]
        if "environment_variables" in conf:
            env_vars = conf["environment_variables"]

        tox_envs[alias] = {
            "python": "python" + python if "py" not in python else python,
            #"deps": deps.split(),
        }
        if "coverage_flags" in conf:
            tox_envs[alias].update(cover=cover)
        if "environment_variables" in conf:
            tox_envs[alias].update(env_vars=env_vars.split())

    for name in os.listdir(os.path.join("ci", "templates")):
        with open(os.path.join(base_path, name), "w") as fh:
            fh.write(jinja.get_template(name).render(tox_environments=tox_envs))
        print("Wrote {}".format(name))
    print("DONE.")



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import, unicode_literals

import datetime as dt
import numpy as np
import pytest
import aacgmv2

class TestDepAACGMV2:
    def setup(self):
        """Runs before every method to create a clean testing setup"""
        self.dtime = dt.datetime(2015, 1, 1, 0, 0, 0)
        self.ddate = dt.date(2015, 1, 1)
        self.lat = None
        self.lon = None

    def teardown(self):
        """Runs after every method to clean up previous testing"""
        del self.dtime, self.ddate, self.lat, self.lon

    @classmethod
    def test_module_structure(self):
        """Test module structure for deprecated routines"""
        assert aacgmv2
        assert aacgmv2.convert
        assert aacgmv2.deprecated.subsol
        assert aacgmv2.deprecated
        assert aacgmv2.deprecated.gc2gd_lat
        assert aacgmv2.deprecated.igrf_dipole_axis

    def test_convert_single_val(self):
        """Test conversion for a single value"""
        self.lat, self.lon = aacgmv2.convert(60, 0, 300, self.dtime)
        assert isinstance(self.lat, np.ndarray)
        assert isinstance(self.lon, np.ndarray)
        assert self.lat.shape == self.lon.shape and self.lat.shape == (1,)
        np.testing.assert_allclose(self.lat, [58.2258], rtol=1e-4)
        np.testing.assert_allclose(self.lon, [81.1685], rtol=1e-4)

    def test_convert_list(self):
        """Test conversion for list input"""
        self.lat, self.lon = aacgmv2.convert([60], [0], [300], self.dtime)
        assert isinstance(self.lat, np.ndarray)
        assert isinstance(self.lon, np.ndarray)
        assert self.lat.shape == self.lon.shape and self.lat.shape == (1,)
        np.testing.assert_allclose(self.lat, [58.2258], rtol=1e-4)
        np.testing.assert_allclose(self.lon, [81.1685], rtol=1e-4)

        self.lat, self.lon = aacgmv2.convert([60, 61], [0, 0], [300, 300],
                                             self.dtime)
        assert isinstance(self.lat, np.ndarray)
        assert isinstance(self.lon, np.ndarray)
        assert self.lat.shape == self.lon.shape and self.lat.shape == (2,)
        np.testing.assert_allclose(self.lat, [58.2258, 59.3186], rtol=1e-4)
        np.testing.assert_allclose(self.lon, [81.1685, 81.6140], rtol=1e-4)

    def test_convert_arr_single(self):
        """Test conversion for array input with one element"""
        self.lat, self.lon = aacgmv2.convert(np.array([60]), np.array([0]),
                                   np.array([300]), self.dtime)
        assert isinstance(self.lat, np.ndarray)
        assert isinstance(self.lon, np.ndarray)
        assert self.lat.shape == self.lon.shape and self.lat.shape == (1,)
        np.testing.assert_allclose(self.lat, [58.2258], rtol=1e-4)
        np.testing.assert_allclose(self.lon, [81.1685], rtol=1e-4)

    def test_convert_arr(self):
        """Test conversion for array input"""
        self.lat, self.lon = aacgmv2.convert(np.array([60, 61]),
                                             np.array([0, 0]),
                                             np.array([300, 300]), self.dtime)
        assert isinstance(self.lat, np.ndarray)
        assert isinstance(self.lon, np.ndarray)
        assert self.lat.shape == self.lon.shape and self.lat.shape == (2,)
        np.testing.assert_allclose(self.lat, [58.2258, 59.3186], rtol=1e-4)
        np.testing.assert_allclose(self.lon, [81.1685, 81.6140], rtol=1e-4)

    def test_convert_list_mix(self):
        """Test conversion for a list and floats"""
        self.lat, self.lon = aacgmv2.convert([60, 61], 0, 300, self.dtime)
        assert isinstance(self.lat, np.ndarray)
        assert isinstance(self.lon, np.ndarray)
        assert self.lat.shape == self.lon.shape and self.lat.shape == (2,)
        np.testing.assert_allclose(self.lat, [58.2258, 59.3186], rtol=1e-4)
        np.testing.assert_allclose(self.lon, [81.1685, 81.6140], rtol=1e-4)

    def test_convert_arr_mix(self):
        """Test conversion for an array and floats"""
        self.lat, self.lon = aacgmv2.convert(np.array([60, 61]), 0, 300,
                                             self.dtime)
        assert isinstance(self.lat, np.ndarray)
        assert isinstance(self.lon, np.ndarray)
        assert self.lat.shape == self.lon.shape and self.lat.shape == (2,)
        np.testing.assert_allclose(self.lat, [58.2258, 59.3186], rtol=1e-4)
        np.testing.assert_allclose(self.lon, [81.1685, 81.6140], rtol=1e-4)

    def test_convert_mult_array_mix(self):
        """Test conversion for a multi-dim array and floats"""
        self.lat, self.lon = aacgmv2.convert(np.array([[60, 61, 62],
                                                       [63, 64, 65]]), 0,
                                             300, self.dtime)
        assert isinstance(self.lat, np.ndarray)
        assert isinstance(self.lon, np.ndarray)
        assert self.lat.shape == self.lon.shape and self.lat.shape == (2, 3)
        np.testing.assert_allclose(self.lat, [[58.2258, 59.3186, 60.4040],
                                         [61.4820, 62.5528, 63.6164]],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.lon, [[81.1685, 81.6140, 82.0872],
                                         [82.5909, 83.1286, 83.7039]],
                                   rtol=1e-4)

    def test_convert_unequal_arra(self):
        """Test conversion for unequal sized arrays"""
        self.lat, self.lon = aacgmv2.convert(np.array([[60, 61, 62],
                                                       [63, 64, 65]]),
                                             np.array([0]), np.array([300]),
                                             self.dtime)
        assert isinstance(self.lat, np.ndarray)
        assert isinstance(self.lon, np.ndarray)
        assert self.lat.shape == self.lon.shape and self.lat.shape == (2, 3)
        np.testing.assert_allclose(self.lat, [[58.2258, 59.3186, 60.4040],
                                              [61.4820, 62.5528, 63.6164]],
                                   rtol=1e-4)
        np.testing.assert_allclose(self.lon, [[81.1685, 81.6140, 82.0872],
                                              [82.5909, 83.1286, 83.7039]],
                                   rtol=1e-4)

    def test_convert_location_failure(self):
        """Test conversion with a bad location"""
        self.lat, self.lon = aacgmv2.convert([0], [0], [0], self.dtime)
        assert isinstance(self.lat, np.ndarray)
        assert isinstance(self.lon, np.ndarray)
        assert self.lat.shape == self.lon.shape and self.lat.shape == (1,)
        assert np.all([np.isnan(self.lat), np.isnan(self.lon)])

    def test_convert_time_failure(self):
        """Test conversion with a bad time"""
        with pytest.raises(AssertionError):
            self.lat, self.lon = aacgmv2.convert([60], [0], [300], None)

    def test_convert_datetime_date(self):
        """Test conversion with date and datetime input"""
        self.lat, self.lon = aacgmv2.convert([60], [0], [300], self.ddate)
        np.testing.assert_allclose(self.lat, [58.2258], rtol=1e-4)
        np.testing.assert_allclose(self.lon, [81.1685], rtol=1e-4)

    def test_warning_below_ground_convert(self):
        """ Test that a warning is issued if altitude is below zero"""
        import logbook
        lwarn = u"conversion not intended for altitudes < 0 km"

        with logbook.TestHandler() as handler:
            self.lat, self.lon = aacgmv2.convert([60], [0], [-1], self.dtime)
            assert handler.has_warning(lwarn)

        handler.close()

    def test_convert_maxalt_failure(self):
        """For an array, test failure for an altitude too high for
        coefficients"""
        with pytest.raises(ValueError):
            aacgmv2.convert([60], [0], [2001], self.dtime)

    def test_convert_lat_failure(self):
        """Test error return for co-latitudes above 90 for an array"""
        with pytest.raises(AssertionError):
            aacgmv2.convert([91, 60, -91], 0, 300, self.dtime)

    def test_subsol(self):
        """Test the subsolar calculation"""
        doy = int(self.dtime.strftime("%j"))
        ut = self.dtime.hour * 3600.0 + self.dtime.minute * 60.0 + \
             self.dtime.second
        self.lon, self.lat = aacgmv2.deprecated.subsol(self.dtime.year, doy, ut)

        np.testing.assert_almost_equal(self.lon, -179.2004, decimal=4)
        np.testing.assert_almost_equal(self.lat, -23.0431, decimal=4)

    def test_gc2gd_lat(self):
        """Test the geocentric to geodetic conversion"""
        self.lat = aacgmv2.deprecated.gc2gd_lat(45.0)

        np.testing.assert_almost_equal(self.lat, 45.1924, decimal=4)

    def test_gc2gd_lat_list(self):
        """Test the geocentric to geodetic conversion"""
        self.lat = [45.0, -45.0]
        self.lat = aacgmv2.deprecated.gc2gd_lat(self.lat)

        np.testing.assert_allclose(self.lat, [45.1924, -45.1924], rtol=1.0e-4)

    def test_gc2gd_lat_arr(self):
        """Test the geocentric to geodetic conversion"""
        self.lat = np.array([45.0, -45.0])
        self.lat = aacgmv2.deprecated.gc2gd_lat(self.lat)

        np.testing.assert_allclose(self.lat, [45.1924, -45.1924], rtol=1.0e-4)

    def test_igrf_dipole_axis(self):
        """Test the IGRF dipole axis calculation"""
        m = aacgmv2.deprecated.igrf_dipole_axis(self.dtime)

        np.testing.assert_allclose(m, [0.050253, -0.160608, 0.985738],
                                   rtol=1.0e-4)



#!/usr/bin/env python
#-*- encoding: utf-8 -*-
from __future__ import absolute_import

import io
import os
import re
from os import path

from setuptools import setup, find_packages
from distutils.core import Extension

def read(fname, **kwargs):
    return io.open(path.join(path.dirname(__file__), fname),
                   encoding=kwargs.get('encoding', 'utf8')).read()

# enable code coverage for C code
# We can't use CFLAGS=-coverage in tox.ini, since that may mess with
# compiling dependencies (e.g. numpy). Therefore we set PY_CCOV=-coverage
# in tox.ini and copy it to CFLAGS here (after deps have been installed)
if 'PY_CCOV' in os.environ.keys():
    os.environ['CFLAGS'] = os.environ['PY_CCOV']


setup(
    name='aacgmv2',
    version='2.5.0',
    license='MIT',
    description='A Python wrapper for AACGM-v2 magnetic coordinates',
    long_description='%s\n%s' % (read('README.rst'),
                                 re.sub(':[a-z]+:`~?(.*?)`',
                                        r'``\1``', read('CHANGELOG.rst'))),
    author='Angeline G. Burrell, Christer van der Meeren',
    author_email='angeline.burrell@nrl.navy.mil',
    url='https://github.com/aburrell/aacgmv2',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list:
        #   http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Utilities',
    ],
    keywords=[
        'aacgm',
        'aacgm-v2',
        'aacgmv2',
        'magnetic coordinates',
        'altitude adjusted corrected geomagnetic coordinates',
        'mlt',
        'magnetic local time',
        'conversion',
        'converting',
    ],
    install_requires=[
        'numpy',
        'logbook',
    ],
    extras_require={'test':['pytest'],
    },
    ext_modules=[
        Extension('aacgmv2._aacgmv2',
                  sources=['aacgmv2/aacgmv2module.c',
                           'c_aacgmv2/src/aacgmlib_v2.c',
                           'c_aacgmv2/src/astalglib.c',
                           'c_aacgmv2/src/genmag.c',
                           'c_aacgmv2/src/igrflib.c',
                           'c_aacgmv2/src/mlt_v2.c',
                           'c_aacgmv2/src/rtime.c'],
                  include_dirs=['c_aacgmv2/include'])
    ],
    entry_points={
        'console_scripts': [
            'aacgmv2 = aacgmv2.__main__:main',
        ]
    },
)
